# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

"""
Access to Memory (DRAM) Serial Presence Detect (SPD) EEPROM

References:

http://www.jedec.org/sites/default/files/docs/4_01_02R19.pdf
http://www.jedec.org/sites/default/files/docs/4_01_02_10R17.pdf
http://www.jedec.org/sites/default/files/docs/4_01_02_11R24.pdf
http://www.jedec.org/sites/default/files/docs/4_01_02_12R23A.pdf
https://www.simmtester.com/News/PublicationArticle/184
https://www.simmtester.com/News/PublicationArticle/153
https://www.simmtester.com/News/PublicationArticle/101
http://en.wikipedia.org/wiki/Serial_presence_detect
"""

import struct
from typing import Any, List
from collections import namedtuple

from chipsec.library.logger import logger, print_buffer_bytes

SPD_SMBUS_ADDRESS = 0xA0  # A2, A4, A6, A8, AA, AC, AE
SPD_SMBUS_ADDRESS_DIMM0 = SPD_SMBUS_ADDRESS
SPD_SMBUS_ADDRESS_DIMM1 = SPD_SMBUS_ADDRESS + 0x2
SPD_SMBUS_ADDRESS_DIMM2 = SPD_SMBUS_ADDRESS + 0x4
SPD_SMBUS_ADDRESS_DIMM3 = SPD_SMBUS_ADDRESS + 0x6
SPD_SMBUS_ADDRESS_DIMM4 = SPD_SMBUS_ADDRESS + 0x8
SPD_SMBUS_ADDRESS_DIMM5 = SPD_SMBUS_ADDRESS + 0xA
SPD_SMBUS_ADDRESS_DIMM6 = SPD_SMBUS_ADDRESS + 0xC
SPD_SMBUS_ADDRESS_DIMM7 = SPD_SMBUS_ADDRESS + 0xE
MAX_DIMM_SPD_COUNT = 8

SPD_DIMMS = {}
for i in range(MAX_DIMM_SPD_COUNT):
    SPD_DIMMS[SPD_SMBUS_ADDRESS + i * 2] = f'DIMM{i:d}'

SPD_DIMM_ADDRESSES = {}
for i in range(MAX_DIMM_SPD_COUNT):
    SPD_DIMM_ADDRESSES[f'DIMM{i:d}'] = SPD_SMBUS_ADDRESS + i * 2

###############################################################################
#
# SPD Decode
#
# References:
# http://www.jedec.org/sites/default/files/docs/4_01_02R19.pdf
# http://www.jedec.org/sites/default/files/docs/4_01_02_10R17.pdf
# http://www.jedec.org/sites/default/files/docs/4_01_02_11R24.pdf
# http://www.jedec.org/sites/default/files/docs/4_01_02_12R23A.pdf
# http://www.simmtester.com/page/news/showpubnews.asp?num=184
# http://www.simmtester.com/page/news/showpubnews.asp?num=153
# http://www.simmtester.com/page/news/showpubnews.asp?num=101
# http://en.wikipedia.org/wiki/Serial_presence_detect
#
# @TODO: add decode of other fields
#
###############################################################################

#
# DDR/DDR2/DDR3/DDR4 SPD
#
SPD_OFFSET_DRAM_DEVICE_TYPE = 2  # Fundamental Memory (DRAM) Type

#
# DDR SPD
#
SPD_OFFSET_DDR_SPD_BYTES = 0
SPD_OFFSET_DDR_SPD_SIZE = 1
SPD_OFFSET_DDR_ROW_ADDRESS_COUNT = 3
SPD_OFFSET_DDR_COL_ADDRESS_COUNT = 4
SPD_OFFSET_DDR_BANKDS_COUNT = 5
SPD_OFFSET_DDR_MODULE_WIDTH_LOW = 6
SPD_OFFSET_DDR_MODULE_WIDTH_HIGH = 7
SPD_OFFSET_DDR_VOLTAGE_IFACE_LEVEL = 8
SPD_OFFSET_DDR_CLOCK_FREQUENCY = 9
SPD_OFFSET_DDR_tAC = 10
SPD_OFFSET_DDR_DIMM_CONFIGURATION_TYPE = 11
SPD_OFFSET_DDR_REFRESH_RATE_TYPE = 12
SPD_OFFSET_DDR_PRIMARY_SDRAM_WIDTH = 13
SPD_OFFSET_DDR_ECC_SDRAM_WIDTH = 14
SPD_OFFSET_DDR_tCCD_MIN = 15

#
# DDR3 SPD
#
SPD_OFFSET_DDR3_SPD_BYTES = 0  # SPD Bytes Written, Device Size, CRC coverage/range
SPD_OFFSET_DDR3_SPD_REVISION = 1  # SPD Revision
SPD_OFFSET_DDR3_MODULE_TYPE = 3  # Module Type
SPD_OFFSET_DDR3_SDRAM_DENSITY_BANKS = 4  # SDRAM Density and Banks
SPD_OFFSET_DDR3_SDRAM_ADDRESSING = 5  # SDRAM Addressing
SPD_OFFSET_DDR3_VDD = 6  # Module Nominal Voltage, VDD
SPD_OFFSET_DDR3_MODULE_ORGANIZATION = 7  # Module Organization
SPD_OFFSET_DDR3_MEMORY_BUS_WIDTH_ECC = 8  # Module Memory Bus Width
SPD_OFFSET_DDR3_FTB = 9  # Fine Time Base (FTB) Divident / Divisor
SPD_OFFSET_DDR3_MTB_DIVIDENT = 10  # Medium Time Base (MTB) Divident
SPD_OFFSET_DDR3_MTB_DIVISOR = 11  # Medium Time Base (MTB) Divisor
SPD_OFFSET_DDR3_tCK_MIN = 12  # SDRAM Minimum Cycle Time (tCKmin)
SPD_OFFSET_DDR3_RESERVED13 = 13  # Reserved
SPD_OFFSET_DDR3_CAS_LATENCY_LOW = 14  # CAS Latencies Supported, LSB
SPD_OFFSET_DDR3_CAS_LATENCY_HIGH = 15  # CAS Latencies Supported, MSB

#
# DDR4 SPD
#
# Base Configuration and DRAM Parameters
SPD_OFFSET_DDR4_SPD_BYTES = 0  # SPD Bytes Written, Device Size, CRC coverage/range
SPD_OFFSET_DDR4_SPD_REVISION = 1  # SPD Revision
SPD_OFFSET_DDR4_MODULE_TYPE = 3  # Module Type
SPD_OFFSET_DDR4_SDRAM_DENSITY_BANKS = 4  # SDRAM Density and Banks
SPD_OFFSET_DDR4_SDRAM_ADDRESSING = 5  # SDRAM Addressing
SPD_OFFSET_DDR4_SDRAM_PACKAGE_TYPE = 6  # SDRAM Package Type
SPD_OFFSET_DDR4_OPTIONAL_FEATURES = 7  # SDRAM Optional Features
SPD_OFFSET_DDR4_THERMAL_AND_REFRESH = 8  # SDRAM Thermal and Refresh Options
SPD_OFFSET_DDR4_OPTIONAL_FEATURES_1 = 9  # Other Optional Features
SPD_OFFSET_DDR4_RESERVED10 = 10  # Reserved (must be 0x00)
SPD_OFFSET_DDR4_VDD = 11  # Module Nominal Voltage, VDD
SPD_OFFSET_DDR4_MODULE_ORGANIZATION = 12  # Module Organization
SPD_OFFSET_DDR4_MEMORY_BUS_WIDTH_ECC = 13  # Module Memory Bus Width
SPD_OFFSET_DDR4_MODULE_THERMAL_SENSOR = 14  # Module Thermal Sensor
SPD_OFFSET_DDR4_MODULE_TYPE_EXTENDED = 15  # Extended Module Type


#
# Fundamental Memory Type
# Ref: http://www.jedec.org/sites/default/files/docs/4_01_02_01R12.pdf
#
DRAM_DEVICE_TYPE_FPM_DRAM = 0x1
DRAM_DEVICE_TYPE_EDO = 0x2
DRAM_DEVICE_TYPE_PIPELINED_NIBBLE = 0x3
DRAM_DEVICE_TYPE_SDR = 0x4
DRAM_DEVICE_TYPE_MULTIPLEXED_ROM = 0x5
DRAM_DEVICE_TYPE_DDR = 0x7
DRAM_DEVICE_TYPE_DDR2 = 0x8
DRAM_DEVICE_TYPE_DDR3 = 0x0B
DRAM_DEVICE_TYPE_DDR4 = 0x0C
DRAM_DEVICE_TYPE = {
    DRAM_DEVICE_TYPE_FPM_DRAM: 'Standard Fast Page Mode DRAM',
    DRAM_DEVICE_TYPE_EDO: 'EDO DRAM',
    DRAM_DEVICE_TYPE_PIPELINED_NIBBLE: 'Pipelined Nibble',
    DRAM_DEVICE_TYPE_SDR: 'Sync DRAM (SDRAM)',
    DRAM_DEVICE_TYPE_MULTIPLEXED_ROM: 'Multiplexed ROM',
    DRAM_DEVICE_TYPE_DDR: 'DDR SDRAM',
    DRAM_DEVICE_TYPE_DDR2: 'DDR2 SDRAM',
    DRAM_DEVICE_TYPE_DDR3: 'DDR3 SDRAM',
    DRAM_DEVICE_TYPE_DDR4: 'DDR4 SDRAM'
}

MODULE_TYPE_UNDEFINED = 0x0
MODULE_TYPE_RDIMM = 0x1
MODULE_TYPE_UDIMM = 0x2
MODULE_TYPE_SODIMM = 0x3
MODULE_TYPE_LRDIMM = 0x4
MODULE_TYPE = {
    MODULE_TYPE_UNDEFINED: 'Undefined',
    MODULE_TYPE_RDIMM: 'Registered Long DIMM',
    MODULE_TYPE_UDIMM: 'Unbuffered Long DIMM',
    MODULE_TYPE_SODIMM: 'Small Outline DIMM',
    MODULE_TYPE_LRDIMM: 'LR-DIMM'
}

SPD_REVISION_0_0 = 0x00
SPD_REVISION_0_7 = 0x07
SPD_REVISION_0_8 = 0x08
SPD_REVISION_0_9 = 0x09
SPD_REVISION_1_0 = 0x10
SPD_REVISION_1_1 = 0x11
SPD_REVISION_1_2 = 0x12
SPD_REVISION_1_3 = 0x13


def SPD_REVISION(revision: int) -> str:
    return (f'{revision >> 4:d}.{revision & 0xF:d}')


def dram_device_type_name(dram_type: int) -> str:
    dt_name = DRAM_DEVICE_TYPE[dram_type] if dram_type in DRAM_DEVICE_TYPE else 'unknown'
    return dt_name


def module_type_name(module_type: int) -> str:
    mt_name = MODULE_TYPE[module_type] if module_type in MODULE_TYPE else 'unknown'
    return mt_name


SPD_DDR_FORMAT = '=4B'


class SPD_DDR(namedtuple('SPD_DDR', 'SPDBytes TotalBytes DeviceType RowAddressCount')):
    __slots__ = ()

    def __str__(self) -> str:
        return f"""------------------------------------------------------------------
SPD DDR
------------------------------------------------------------------
[0] Number of SPD bytes written: 0x{self.SPDBytes:02X}
[1] Total number of bytes      : 0x{self.TotalBytes:02X}
[2] DRAM Memory Type           : 0x{self.DeviceType:02X} ({dram_device_type_name(self.DeviceType)})
[3] Number of Row Addresses    : 0x{self.RowAddressCount:02X}
------------------------------------------------------------------
"""


SPD_DDR2_FORMAT = '=4B'


class SPD_DDR2(namedtuple('SPD_DDR2', 'SPDBytes TotalBytes DeviceType RowAddressCount')):
    __slots__ = ()

    def __str__(self) -> str:
        return f"""------------------------------------------------------------------
SPD DDR2
------------------------------------------------------------------
[0] Number of SPD bytes written: 0x{self.SPDBytes:02X}
[1] Total number of bytes      : 0x{self.TotalBytes:02X}
[2] DRAM Memory Type           : 0x{self.DeviceType:02X} ({dram_device_type_name(self.DeviceType)})
[3] Number of Row Addresses    : 0x{self.RowAddressCount:02X}
------------------------------------------------------------------
"""


SPD_DDR3_FORMAT = '=16B'


class SPD_DDR3(namedtuple('SPD_DDR3', 'SPDBytes Revision DeviceType ModuleType ChipSize Addressing Voltages ModuleOrg BusWidthECC FTB MTBDivident MTBDivisor tCKMin RsvdD CASLo CASHi')):
    __slots__ = ()

    def __str__(self) -> str:
        return f"""------------------------------------------------------------------
SPD DDR3
------------------------------------------------------------------
[0x00] SPD Bytes Written, Device Size, CRC: 0x{self.SPDBytes:02X}
[0x01] SPD Revision                       : 0x{self.Revision:02X} ({SPD_REVISION(self.Revision)})
[0x02] DRAM Memory Type                   : 0x{self.DeviceType:02X} ({dram_device_type_name(self.DeviceType)})
[0x03] Module Type                        : 0x{self.ModuleType:02X} ({module_type_name(self.ModuleType)})
[0x04] SDRAM Density and Banks            : 0x{self.ChipSize:02X}
[0x05] SDRAM Addressing (Row/Column Bits) : 0x{self.Addressing:02X}
[0x06] Module Nominal Voltage, VDD        : 0x{self.Voltages:02X}
[0x07] Module Organization                : 0x{self.ModuleOrg:02X}
[0x08] Module Memory Bus Width, ECC       : 0x{self.BusWidthECC:02X}
[0x09] FTB Divident/Divisor               : 0x{self.FTB:02X}
[0x0A] MTB Divident                       : 0x{self.MTBDivident:02X}
[0x0B] MTB Divisor                        : 0x{self.MTBDivisor:02X}
[0x0C] SDRAM Minimum Cycle Time (tCKmin)  : 0x{self.tCKMin:02X}
[0x0D] Reserved                           : 0x{self.RsvdD:02X}
[0x0E] CAS Latencies Supported (LSB)      : 0x{self.CASLo:02X}
[0x0F] CAS Latencies Supported (MSB)      : 0x{self.CASHi:02X}
------------------------------------------------------------------
"""


SPD_DDR4_FORMAT = '=16B'


class SPD_DDR4(namedtuple('SPD_DDR4', 'SPDBytes Revision DeviceType ModuleType Density Addressing PackageType OptFeatures ThermalRefresh OptFeatures1 ReservedA VDD ModuleOrg BusWidthECC ThermSensor ModuleTypeExt')):
    __slots__ = ()

    def __str__(self) -> str:
        return f"""------------------------------------------------------------------
SPD DDR4
------------------------------------------------------------------
Base Configuration and DRAM Parameters
[0x00] SPD Bytes Written, Device Size, CRC: 0x{self.SPDBytes:02X}
[0x01] SPD Revision                       : 0x{self.Revision:02X} ({SPD_REVISION(self.Revision)})
[0x02] DRAM Memory Type                   : 0x{self.DeviceType:02X} ({dram_device_type_name(self.DeviceType)})
[0x03] Module Type                        : 0x{self.ModuleType:02X} ({module_type_name(self.ModuleType)})
[0x04] SDRAM Density and Banks            : 0x{self.Density:02X}
[0x05] SDRAM Addressing (Row/Column Bits) : 0x{self.Addressing:02X}
[0x06] SDRAM Package Type                 : 0x{self.PackageType:02X}
[0x07] SDRAM Optional Features            : 0x{self.OptFeatures:02X}
[0x08] SDRAM Thermal and Refresh Options  : 0x{self.ThermalRefresh:02X}
[0x09] Other Optional Features            : 0x{self.OptFeatures1:02X}
[0x0A] Reserved (== 0x00)                 : 0x{self.ReservedA:02X}
[0x0B] Module Nominal Voltage, VDD        : 0x{self.VDD:02X}
[0x0C] Module Organization                : 0x{self.ModuleOrg:02X}
[0x0D] Module Memory Bus Width            : 0x{self.BusWidthECC:02X}
[0x0E] Module Thermal Sensor              : 0x{self.ThermSensor:02X}
[0x0F] Extended Module Type               : 0x{self.ModuleTypeExt:02X}
------------------------------------------------------------------
"""


###############################################################################
#
# Main SPD HAL component class
#
###############################################################################

class SPD:
    def __init__(self, smbus):
        self.smbus = smbus

    def read_byte(self, offset: int, device: int = SPD_SMBUS_ADDRESS) -> int:
        return self.smbus.read_byte(device, offset)

    def write_byte(self, offset: int, value: int, device: int = SPD_SMBUS_ADDRESS) -> bool:
        return self.smbus.write_byte(device, offset, value)

    def read_range(self, start_offset: int, size: int, device: int = SPD_SMBUS_ADDRESS) -> bytes:
        return bytes(self.read_byte(start_offset + i, device) for i in range(size))

    def write_range(self, start_offset: int, buffer: bytes, device: int = SPD_SMBUS_ADDRESS) -> bool:
        for i, b in enumerate(buffer):
            self.write_byte(start_offset + i, b, device)
        return True

    def dump_spd_rom(self, device: int = SPD_SMBUS_ADDRESS) -> bytes:
        return self.read_range(0x0, 0x100, device)

    #
    # Decoding certain bytes of DIMM SPD: may be dependent on the DRAM type
    #
    def getDRAMDeviceType(self, device: int = SPD_SMBUS_ADDRESS) -> int:
        dram_type = self.read_byte(SPD_OFFSET_DRAM_DEVICE_TYPE, device)
        logger().log_hal(f'[spd][0x{device:02X}] DRAM Device Type (byte 2): 0x{dram_type:01X}')
        return dram_type

    def getModuleType(self, device: int = SPD_SMBUS_ADDRESS) -> int:
        module_type = self.read_byte(SPD_OFFSET_DDR3_MODULE_TYPE, device)
        logger().log_hal(f'[spd][0x{device:02X}] Module Type (byte 3): 0x{module_type:01X}')
        return module_type

    def isECC(self, device: int = SPD_SMBUS_ADDRESS) -> bool:
        device_type = self.getDRAMDeviceType(device)
        ecc_supported = False
        ecc_off = 0
        ecc = None
        if DRAM_DEVICE_TYPE_DDR3 == device_type:
            ecc_off = SPD_OFFSET_DDR3_MEMORY_BUS_WIDTH_ECC
            ecc = self.read_byte(ecc_off, device)
            ecc_supported = (0xB == ecc)
        elif DRAM_DEVICE_TYPE_DDR4 == device_type:
            ecc_off = SPD_OFFSET_DDR4_MEMORY_BUS_WIDTH_ECC
            ecc = self.read_byte(ecc_off, device)
            ecc_supported = (0xB == ecc)
        elif DRAM_DEVICE_TYPE_DDR == device_type or DRAM_DEVICE_TYPE_DDR2 == device_type:
            ecc_off = SPD_OFFSET_DDR_DIMM_CONFIGURATION_TYPE
            ecc = self.read_byte(ecc_off, device)
            ecc_supported = (0x2 == ecc)
            ecc_width = self.read_byte(SPD_OFFSET_DDR_ECC_SDRAM_WIDTH, device)
            logger().log_hal(f'[spd][0x{device:02X}] DDR/DDR2 ECC width (byte {SPD_OFFSET_DDR_ECC_SDRAM_WIDTH:d}): 0x{ecc_width:02X}')

        if logger().HAL:
            if ecc is None:
                logger().log(f'[spd][0x{device:02X}] Unable to determine ECC support')
            else:
                not_str = '' if ecc_supported else 'not '
                logger().log(f'[spd][0x{device:02X}] ECC is {not_str}supported by the DIMM (byte {ecc_off:d} = 0x{ecc:02X})')
        return ecc_supported

    def detect(self) -> List[int]:
        _dimms = []
        for d in SPD_DIMMS:
            if self.isSPDPresent(d):
                _dimms.append(d)
        if logger().HAL:
            logger().log('Detected the following SPD devices:')
            for _dimm in _dimms:
                logger().log(f"{SPD_DIMMS[_dimm]}: 0x{_dimm:02X}")
        return _dimms

    def isSPDPresent(self, device: int = SPD_SMBUS_ADDRESS) -> bool:
        device_type = self.getDRAMDeviceType(device)
        is_spd_present = (device_type != 0xFF)
        not_str = '' if is_spd_present else 'not '
        logger().log_hal(f'[spd][0x{device:02X}] Detecting SPD.. {not_str}found (DRAM memory type = 0x{device_type:X})')
        return is_spd_present

    def decode(self, device: int = SPD_SMBUS_ADDRESS) -> None:
        spd: Any = None
        device_type = self.getDRAMDeviceType(device)
        spd_buffer = self.dump_spd_rom(device)
        logger().log(f'[spd][0x{device:02X}] Serial Presence Detect (SPD) EEPROM contents:')
        print_buffer_bytes(spd_buffer)

        if DRAM_DEVICE_TYPE_DDR == device_type:
            spd = SPD_DDR(*struct.unpack_from(SPD_DDR_FORMAT, spd_buffer))
        elif DRAM_DEVICE_TYPE_DDR2 == device_type:
            spd = SPD_DDR2(*struct.unpack_from(SPD_DDR2_FORMAT, spd_buffer))
        elif DRAM_DEVICE_TYPE_DDR3 == device_type:
            spd = SPD_DDR3(*struct.unpack_from(SPD_DDR3_FORMAT, spd_buffer))
        elif DRAM_DEVICE_TYPE_DDR4 == device_type:
            spd = SPD_DDR4(*struct.unpack_from(SPD_DDR4_FORMAT, spd_buffer))
        else:
            logger().log_warning('[spd] Unsupported SPD format')

        if spd is not None:
            logger().log(str(spd))



# -*- coding: utf-8 -*-
# # Copyright (c) 2020 Intel Corporation
# SPDX-License-Identifier: GPL-2.0-only
#
# This file incorporates work covered by the following copyright and
# permission notice:
#
# Copyright (c) 2014 Anders Høst
# Copyright (c) 2018 Anders Høst
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import mmap
from collections.abc import Callable
from ctypes import CDLL, CFUNCTYPE, addressof, c_uint16, c_uint32, c_void_p, get_errno

from chipsec.library.exceptions import OsHelperError

IN_PORT = bytes((
    0x55,                   # push   %rbp
    0x48, 0x89, 0xe5,       # mov    %rsp,%rbp
    0x89, 0xf8,             # mov    %edi,%eax
    0x66, 0x89, 0x45, 0xec,  # mov    %ax,-0x14(%rbp)
    0x0f, 0xb7, 0x45, 0xec,  # movzwl -0x14(%rbp),%eax
    0x89, 0xc2,             # mov    %eax,%edx
    0xed,                   # in     (%dx),%eax
    0x89, 0x45, 0xfc,       # mov    %eax,-0x4(%rbp)
    0x8b, 0x45, 0xfc,       # mov    -0x4(%rbp),%eax
    0x5d,                   # pop    %rbp
    0xc3,                   # retq
))


OUT_PORT = bytes((
    0x55,                   # push   %rbp
    0x48, 0x89, 0xe5,       # mov    %rsp,%rbp
    0x89, 0x7d, 0xfc,       # mov    %edi,-0x4(%rbp)
    0x89, 0xf0,             # mov    %esi,%eax
    0x66, 0x89, 0x45, 0xf8,  # mov    %ax,-0x8(%rbp)
    0x8b, 0x45, 0xfc,       # mov    -0x4(%rbp),%eax
    0x0f, 0xb7, 0x55, 0xf8,  # movzwl -0x8(%rbp),%edx
    0xef,                   # out    %eax,(%dx)
    0x90,                   # nop
    0x5d,                   # pop    %rbp
    0xc3,                   # retq
))


class Ports:
    # Use a unique Ports instance, to avoid allocating memory mappings every time it is used
    instance = None

    def __init__(self) -> None:
        clib = CDLL("libc.so.6", use_errno=True)
        if clib.iopl(3) == -1:
            raise OsHelperError("Unable to use I/O ports using iopl", get_errno())

        self.inl_addr = mmap.mmap(-1, mmap.PAGESIZE, flags=mmap.MAP_PRIVATE, prot=mmap.PROT_READ | mmap.PROT_WRITE | mmap.PROT_EXEC)
        self.inl_addr.write(IN_PORT)
        in_func_type = CFUNCTYPE(c_uint32, c_uint16)
        in_fp = c_void_p.from_buffer(self.inl_addr)
        self.inl_ptr: Callable[[int], int] = in_func_type(addressof(in_fp))

        self.outl_addr = mmap.mmap(-1, mmap.PAGESIZE, flags=mmap.MAP_PRIVATE, prot=mmap.PROT_READ | mmap.PROT_WRITE | mmap.PROT_EXEC)
        self.outl_addr.write(OUT_PORT)
        out_func_type = CFUNCTYPE(None, c_uint32, c_uint16)
        out_fp = c_void_p.from_buffer(self.outl_addr)
        self.outl_ptr: Callable[[int, int], None] = out_func_type(addressof(out_fp))

    @classmethod
    def get_instance(cls) -> "Ports":
        if cls.instance is None:
            cls.instance = cls()
        return cls.instance

    def inl(self, port: int) -> int:
        x = self.inl_ptr(port)
        return x

    def outl(self, value: int, port: int) -> None:
        self.outl_ptr(value, port)


class LegacyPci:
    @staticmethod
    def read_pci_config(bus: int, dev: int, func: int, offset: int) -> int:
        ports = Ports.get_instance()
        ports.outl(0x80000000 | (bus << 16) | (dev << 11) | (func << 8) | offset, 0xcf8)
        v = ports.inl(0xcfc)
        return v

    @staticmethod
    def write_pci_config(bus: int, dev: int, func: int, offset: int, value: int) -> None:
        ports = Ports.get_instance()
        ports.outl(0x80000000 | (bus << 16) | (dev << 11) | (func << 8) | offset, 0xcf8)
        ports.outl(value, 0xcfc)



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
# Copyright (c) 2025, Oracle and/or its affiliates.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

"""
Linux helper
"""

import array
import ctypes
import errno
import fcntl
import os
import platform
import struct
import subprocess
import sys
from collections.abc import Iterable
from typing import Dict, List, Optional, Tuple, TYPE_CHECKING
if TYPE_CHECKING:
    from chipsec.library.types import EfiVariableType
    from ctypes import Array

from chipsec.library import defines
from chipsec.helper.oshelper import get_tools_path
from chipsec.library.exceptions import OsHelperError, UnimplementedAPIError
from chipsec.helper.basehelper import Helper
from chipsec.library.logger import logger
import chipsec.library.file

MSGBUS_MDR_IN_MASK = 0x1
MSGBUS_MDR_OUT_MASK = 0x2

IOCTL_BASE = 0x0
IOCTL_RDIO = 0x1
IOCTL_WRIO = 0x2
IOCTL_RDPCI = 0x3
IOCTL_WRPCI = 0x4
IOCTL_RDMSR = 0x5
IOCTL_WRMSR = 0x6
IOCTL_CPUID = 0x7
IOCTL_GET_CPU_DESCRIPTOR_TABLE = 0x8
IOCTL_HYPERCALL = 0x9
IOCTL_SWSMI = 0xA
IOCTL_LOAD_UCODE_PATCH = 0xB
IOCTL_ALLOC_PHYSMEM = 0xC
IOCTL_GET_EFIVAR = 0xD
IOCTL_SET_EFIVAR = 0xE
IOCTL_RDCR = 0x10
IOCTL_WRCR = 0x11
IOCTL_RDMMIO = 0x12
IOCTL_WRMMIO = 0x13
IOCTL_VA2PA = 0x14
IOCTL_MSGBUS_SEND_MESSAGE = 0x15
IOCTL_FREE_PHYSMEM = 0x16
IOCTL_SWSMI_TIMED = 0x17

_tools = {}


class LinuxHelper(Helper):

    DEVICE_NAME = '/dev/chipsec'
    DEV_MEM = '/dev/mem'
    DEV_PORT = '/dev/port'
    MODULE_NAME = 'chipsec'
    SUPPORT_KERNEL26_GET_PAGE_IS_RAM = False
    SUPPORT_KERNEL26_GET_PHYS_MEM_ACCESS_PROT = False
    DKMS_DIR = '/var/lib/dkms/'

    def __init__(self):
        super(LinuxHelper, self).__init__()
        self.os_system = platform.system()
        self.os_release = platform.release()
        self.os_version = platform.version()
        self.os_machine = platform.machine()
        self.os_uname = platform.uname()
        self.name = 'LinuxHelper'
        self.dev_fh = None
        self.dev_mem = None
        self.dev_port = None
        self.dev_msr = None

###############################################################################################
# Driver/service management functions
###############################################################################################

    def get_dkms_module_location(self) -> str:
        version = defines.get_version()
        from os import listdir
        from os.path import isdir, join
        p = os.path.join(self.DKMS_DIR, self.MODULE_NAME, version, self.os_release)
        os_machine_dir_name = [f for f in listdir(p) if isdir(join(p, f))][0]
        return os.path.join(self.DKMS_DIR, self.MODULE_NAME, version, self.os_release, os_machine_dir_name, 'module', 'chipsec.ko')

    def load_chipsec_module(self):
        page_is_ram = ''
        phys_mem_access_prot = ''
        a1 = ''
        a2 = ''
        if self.SUPPORT_KERNEL26_GET_PAGE_IS_RAM:
            page_is_ram = self.get_page_is_ram()
            if not page_is_ram:
                logger().log_debug("Cannot find symbol 'page_is_ram'")
            else:
                a1 = f'a1=0x{page_is_ram}'
        if self.SUPPORT_KERNEL26_GET_PHYS_MEM_ACCESS_PROT:
            phys_mem_access_prot = self.get_phys_mem_access_prot()
            if not phys_mem_access_prot:
                logger().log_debug("Cannot find symbol 'phys_mem_access_prot'")
            else:
                a2 = f'a2=0x{phys_mem_access_prot}'

        driver_path = os.path.join(chipsec.library.file.get_main_dir(), 'chipsec', 'helper', 'linux', 'chipsec.ko')
        if not os.path.exists(driver_path):
            driver_path += '.xz'
            if not os.path.exists(driver_path):
                try:
                    driver_path = self.get_dkms_module_location()
                except Exception:
                    pass
                if not os.path.exists(driver_path):
                    driver_path += '.xz'
                    if not os.path.exists(driver_path):
                        raise Exception('Cannot find chipsec.ko module')
        try:
            subprocess.check_output(['insmod', driver_path, a1, a2])
        except Exception as err:
            raise Exception(f'Could not start Linux Helper, are you running as Admin/root?\n\t{err}')
        uid = gid = 0
        os.chown(self.DEVICE_NAME, uid, gid)
        os.chmod(self.DEVICE_NAME, 600)
        if os.path.exists(self.DEVICE_NAME):
            logger().log_debug(f'Module {self.DEVICE_NAME} loaded successfully')
        else:
            logger().log_error(f'Fail to load module: {driver_path}')
        self.driverpath = f'({driver_path})'

    def unload_chipsec_module(self) -> None:
        if self.driver_loaded or os.path.exists(self.DEVICE_NAME):
            subprocess.call(['rmmod', self.MODULE_NAME])
            logger().log_debug(f'Module for {self.DEVICE_NAME} unloaded successfully')

    def create(self):
        logger().log_debug('[helper] Linux Helper created')
        return True

    def start(self) -> bool:
        self.unload_chipsec_module()
        self.load_chipsec_module()
        self.init()
        logger().log_debug('[helper] Linux Helper started/loaded')
        return True

    def stop(self) -> bool:
        self.close()
        self.unload_chipsec_module()
        logger().log_debug('[helper] Linux Helper stopped/unloaded')
        return True

    def delete(self) -> bool:
        logger().log_debug('[helper] Linux Helper deleted')
        return True

    def init(self) -> None:
        x64 = True if sys.maxsize > 2**32 else False
        self._pack = 'Q' if x64 else 'I'

        estr = 'Unable to open chipsec device. Did you run as root/sudo and load the driver?\n {}'
        try:
            # Do not buffer access to physical memory...
            self.dev_fh = open(self.DEVICE_NAME, 'rb+', buffering=0)
            self.driver_loaded = True
        except IOError as e:
            raise OsHelperError(estr.format(str(e)), e.errno)
        except BaseException as be:
            raise OsHelperError(estr.format(str(be)), errno.ENXIO)
        self._ioctl_base = self.compute_ioctlbase()

    def close(self) -> None:
        if self.dev_fh:
            self.dev_fh.close()
        self.dev_fh = None
        if self.dev_mem:
            os.close(self.dev_mem)
        self.dev_mem = None

    # code taken from /include/uapi/asm-generic/ioctl.h
    # by default itype is 'C' see drivers/linux/include/chipsec.h
    # currently all chipsec ioctl functions are _IOWR
    # currently all size are pointer
    def compute_ioctlbase(self, itype: str = 'C') -> int:
        # define _IOWR(type,nr,size)     _IOC(_IOC_READ|_IOC_WRITE,(type),(nr),(_IOC_TYPECHECK(size)))
        # define _IOC(dir,type,nr,size) \
        #    (((dir)  << _IOC_DIRSHIFT) | \
        #    ((type) << _IOC_TYPESHIFT) | \
        #    ((nr)   << _IOC_NRSHIFT) | \
        #    ((size) << _IOC_SIZESHIFT))
        # IOC_READ | _IOC_WRITE is 3
        # default _IOC_DIRSHIFT is 30
        # default _IOC_TYPESHIFT is 8
        # nr will be 0
        # _IOC_SIZESHIFT is 16
        return (3 << 30) | (ord(itype) << 8) | (struct.calcsize(self._pack) << 16)

    def ioctl(self, nr: int, args: Iterable, *mutate_flag: bool) -> bytes:
        return fcntl.ioctl(self.dev_fh, self._ioctl_base + nr, args)

###############################################################################################
# Actual API functions to access HW resources
###############################################################################################

    def map_io_space(self, base: int, size: int, cache_type: int) -> None:
        raise UnimplementedAPIError('map_io_space')

    def __mem_block(self, sz: int, newval: Optional[bytes] = None) -> bytes:
        if self.dev_fh is not None:
            if newval is None:
                return self.dev_fh.read(sz)
            else:
                res = self.dev_fh.write(newval)
                self.dev_fh.flush()
                return res.to_bytes(4, 'little')
        return b''

    def write_phys_mem(self, phys_address: int, length: int, newval: bytes) -> int:
        if (newval is None) or (self.dev_fh is None):
            return 0
        self.dev_fh.seek(phys_address)
        res = self.__mem_block(length, newval)
        return int.from_bytes(res, 'little')

    def read_phys_mem(self, phys_address: int, length: int) -> bytes:
        self.dev_fh.seek(phys_address)
        return self.__mem_block(length)

    def va2pa(self, va: int) -> Tuple[Optional[int], int]:
        error_code = 0

        in_buf = struct.pack(self._pack, va)
        try:
            out_buf = self.ioctl(IOCTL_VA2PA, in_buf)
            pa = struct.unpack(self._pack, out_buf)[0]
        except IOError as err:
            if logger().DEBUG:
                logger().log_error(f'[helper] Error in va2pa: getting PA for VA 0x{va:016X} failed with IOError: {err.strerror}')
            return (None, err.errno)

        # Check if PA > max physical address
        max_pa = self.cpuid(0x80000008, 0x0)[0] & 0xFF
        if pa > 1 << max_pa:
            if logger().DEBUG:
                logger().log_error(f'[helper] Error in va2pa: PA higher that max physical address: VA (0x{va:016X}) -> PA (0x{pa:016X})')
            error_code = 1
        return (pa, error_code)

    def read_pci_reg(self, bus: int, device: int, function: int, offset: int, size: int = 4) -> int:
        _PCI_DOM = 0  # Change PCI domain, if there is more than one.
        d = struct.pack(f'5{self._pack}', ((_PCI_DOM << 16) | bus), ((device << 16) | function), offset, size, 0)
        try:
            ret = self.ioctl(IOCTL_RDPCI, d)
        except IOError:
            if logger().DEBUG:
                logger().log_error('IOError\n')
            return 0
        x = struct.unpack(f'5{self._pack}', ret)
        return x[4]

    def write_pci_reg(self, bus: int, device: int, function: int, offset: int, value: int, size: int = 4) -> int:
        _PCI_DOM = 0  # Change PCI domain, if there is more than one.
        d = struct.pack(f'5{self._pack}', ((_PCI_DOM << 16) | bus), ((device << 16) | function), offset, size, value)
        try:
            ret = self.ioctl(IOCTL_WRPCI, d)
        except IOError:
            if logger().DEBUG:
                logger().log_error('IOError\n')
            return 0
        x = struct.unpack(f'5{self._pack}', ret)
        return x[4]

    def load_ucode_update(self, cpu_thread_id: int, ucode_update_buf: bytes) -> bool:
        cpu_ucode_thread_id = ctypes.c_int(cpu_thread_id)

        in_buf = struct.pack('=BH', cpu_thread_id, len(ucode_update_buf)) + ucode_update_buf
        in_buf_final = array.array('c', in_buf)
        out_length = 0
        try:
            out_buf = self.ioctl(IOCTL_LOAD_UCODE_PATCH, in_buf_final)
        except IOError:
            if logger().DEBUG:
                logger().log_error('IOError IOCTL Load Patch\n')
            return False

        return True

    def read_io_port(self, io_port: int, size: int) -> int:
        in_buf = struct.pack(f'3{self._pack}', io_port, size, 0)
        out_buf = self.ioctl(IOCTL_RDIO, in_buf)
        try:
            if 1 == size:
                value = struct.unpack(f'3{self._pack}', out_buf)[2] & 0xff
            elif 2 == size:
                value = struct.unpack(f'3{self._pack}', out_buf)[2] & 0xffff
            else:
                value = struct.unpack(f'3{self._pack}', out_buf)[2] & 0xffffffff
        except struct.error:
            if logger().DEBUG:
                logger().log_error(f"DeviceIoControl did not return value of proper size {size:x} (value = '{out_buf}'): returning 0")
            value = 0

        return value

    def write_io_port(self, io_port: int, value: int, size: int) -> bytes:
        in_buf = struct.pack(f'3{self._pack}', io_port, size, value)
        return self.ioctl(IOCTL_WRIO, in_buf)

    def read_cr(self, cpu_thread_id: int, cr_number: int) -> int:
        self.set_affinity(cpu_thread_id)
        cr = 0
        in_buf = struct.pack(f'3{self._pack}', cpu_thread_id, cr_number, cr)
        unbuf = struct.unpack(f'3{self._pack}', self.ioctl(IOCTL_RDCR, in_buf))
        return (unbuf[2])

    def write_cr(self, cpu_thread_id: int, cr_number: int, value: int):
        self.set_affinity(cpu_thread_id)
        in_buf = struct.pack(f'3{self._pack}', cpu_thread_id, cr_number, value)
        self.ioctl(IOCTL_WRCR, in_buf)
        return

    def read_msr(self, thread_id: int, msr_addr: int) -> Tuple[int, int]:
        self.set_affinity(thread_id)
        edx = eax = 0
        in_buf = struct.pack(f'4{self._pack}', thread_id, msr_addr, edx, eax)
        unbuf = struct.unpack(f'4{self._pack}', self.ioctl(IOCTL_RDMSR, in_buf))
        return (unbuf[3], unbuf[2])

    def write_msr(self, thread_id: int, msr_addr: int, eax: int, edx: int):
        self.set_affinity(thread_id)
        in_buf = struct.pack(f'4{self._pack}', thread_id, msr_addr, edx, eax)
        self.ioctl(IOCTL_WRMSR, in_buf)
        return

    def get_descriptor_table(self, cpu_thread_id: int, desc_table_code: int) -> Tuple[int, int, int]:
        self.set_affinity(cpu_thread_id)
        in_buf = struct.pack(f'5{self._pack}', cpu_thread_id, desc_table_code, 0, 0, 0)
        out_buf = self.ioctl(IOCTL_GET_CPU_DESCRIPTOR_TABLE, in_buf)
        (limit, base_hi, base_lo, pa_hi, pa_lo) = struct.unpack(f'5{self._pack}', out_buf)
        pa = (pa_hi << 32) + pa_lo
        base = (base_hi << 32) + base_lo
        return (limit, base, pa)

    def cpuid(self, eax: int, ecx: int) -> Tuple[int, int, int, int]:
        in_buf = struct.pack(f'4{self._pack}', eax, 0, ecx, 0)
        out_buf = self.ioctl(IOCTL_CPUID, in_buf)
        return struct.unpack(f'4{self._pack}', out_buf)

    def alloc_phys_mem(self, num_bytes: int, max_addr: int):
        in_buf = struct.pack('2' + self._pack, num_bytes, max_addr)
        out_buf = self.ioctl(IOCTL_ALLOC_PHYSMEM, in_buf)
        return struct.unpack(f'2{self._pack}', out_buf)

    def free_phys_mem(self, physmem: int):
        in_buf = struct.pack(f'1{self._pack}', physmem)
        out_buf = self.ioctl(IOCTL_FREE_PHYSMEM, in_buf)
        return struct.unpack(f'1{self._pack}', out_buf)[0]

    def read_mmio_reg(self, phys_address: int, size: int) -> int:
        in_buf = struct.pack(f'2{self._pack}', phys_address, size)
        out_buf = self.ioctl(IOCTL_RDMMIO, in_buf)
        reg = out_buf[:size]
        return defines.unpack1(reg, size)

    def write_mmio_reg(self, phys_address: int, size: int, value: int):
        in_buf = struct.pack(f'3{self._pack}', phys_address, size, value)
        out_buf = self.ioctl(IOCTL_WRMMIO, in_buf)

    def get_ACPI_table(self, table_name: str) -> Optional['Array']:
        raise UnimplementedAPIError('get_ACPI_table')

    def enum_ACPI_tables(self) -> Optional['Array']:
        raise UnimplementedAPIError('enum_ACPI_table')

    def msgbus_send_read_message(self, mcr: int, mcrx: int) -> Optional[int]:
        return self.msgbus_send_message(mcr, mcrx)

    def msgbus_send_write_message(self, mcr: int, mcrx: int, mdr: int) -> None:
        self.msgbus_send_message(mcr, mcrx, mdr)
        return None

    def msgbus_send_message(self, mcr: int, mcrx: int, mdr: Optional[int] = None) -> int:
        mdr_out = 0
        if mdr is None:
            in_buf = struct.pack(f'5{self._pack}', MSGBUS_MDR_OUT_MASK, mcr, mcrx, 0, mdr_out)
        else:
            in_buf = struct.pack(f'5{self._pack}', (MSGBUS_MDR_IN_MASK | MSGBUS_MDR_OUT_MASK), mcr, mcrx, mdr, mdr_out)
        out_buf = self.ioctl(IOCTL_MSGBUS_SEND_MESSAGE, in_buf)
        mdr_out = struct.unpack(f'5{self._pack}', out_buf)[4]
        return mdr_out

    def get_affinity(self) -> Optional[int]:
        try:
            affinity = os.sched_getaffinity(0)
            return list(affinity)[0]
        except Exception:
            return None

    def set_affinity(self, thread_id: int) -> Optional[int]:
        try:
            os.sched_setaffinity(os.getpid(), {thread_id})
            return thread_id
        except Exception:
            return None

    #########################################################
    # (U)EFI Variable API
    #########################################################

    def EFI_supported(self) -> bool:
        return os.path.exists('/sys/firmware/efi/vars/') or os.path.exists('/sys/firmware/efi/efivars/')

    def delete_EFI_variable(self, name: str, guid: str) -> int:
        return self.kern_set_EFI_variable(name, guid, b'')

    def list_EFI_variables(self) -> Optional[Dict[str, List['EfiVariableType']]]:
        return self.kern_list_EFI_variables()

    def get_EFI_variable(self, name: str, guid: str, attrs: Optional[int] = None) -> bytes:
        return self.kern_get_EFI_variable(name, guid)

    def set_EFI_variable(self, name: str, guid: str, buffer: bytes, buffer_size: int, attrs: Optional[int] = None) -> int:
        return self.kern_set_EFI_variable(name, guid, buffer, attrs)

    #
    # Internal (U)EFI Variable API functions via CHIPSEC kernel module
    #

    def kern_get_EFI_variable_full(self, name: str, guid: str) -> 'EfiVariableType':
        status_dict = {
            0: 'EFI_SUCCESS',
            1: 'EFI_LOAD_ERROR',
            2: 'EFI_INVALID_PARAMETER',
            3: 'EFI_UNSUPPORTED',
            4: 'EFI_BAD_BUFFER_SIZE',
            5: 'EFI_BUFFER_TOO_SMALL',
            6: 'EFI_NOT_READY',
            7: 'EFI_DEVICE_ERROR',
            8: 'EFI_WRITE_PROTECTED',
            9: 'EFI_OUT_OF_RESOURCES',
            14: 'EFI_NOT_FOUND',
            26: 'EFI_SECURITY_VIOLATION'
        }
        off = 0
        data = b''
        attr = 0
        buf = b''
        hdr = 0
        base = 12
        namelen = len(name)
        header_size = 52
        data_size = header_size + namelen
        guid0 = int(guid[:8], 16)
        guid1 = int(guid[9:13], 16)
        guid2 = int(guid[14:18], 16)
        guid3 = int(guid[19:21], 16)
        guid4 = int(guid[21:23], 16)
        guid5 = int(guid[24:26], 16)
        guid6 = int(guid[26:28], 16)
        guid7 = int(guid[28:30], 16)
        guid8 = int(guid[30:32], 16)
        guid9 = int(guid[32:34], 16)
        guid10 = int(guid[34:], 16)

        in_buf = struct.pack(f'13I{str(namelen)}s', data_size, guid0, guid1, guid2, guid3, guid4, guid5, guid6, guid7, guid8, guid9, guid10, namelen, name.encode())
        buffer = array.array('B', in_buf)
        stat = self.ioctl(IOCTL_GET_EFIVAR, buffer)
        new_size, status = struct.unpack('2I', buffer[:8])

        if (status == 0x5):
            data_size = new_size + header_size + namelen  # size sent by driver + size of header (size + guid) + size of name
            in_buf = struct.pack(f'13I{str(namelen + new_size)}s', data_size, guid0, guid1, guid2, guid3, guid4, guid5, guid6, guid7, guid8, guid9, guid10, namelen, name.encode())
            buffer = array.array('B', in_buf)
            try:
                stat = self.ioctl(IOCTL_GET_EFIVAR, buffer)
            except IOError:
                if logger().DEBUG:
                    logger().log_error('IOError IOCTL GetUEFIvar\n')
                return (off, buf, hdr, b'', guid, attr)
            new_size, status = struct.unpack('2I', buffer[:8])

        if (new_size > data_size):
            if logger().DEBUG:
                logger().log_error('Incorrect size returned from driver')
            return (off, buf, hdr, b'', guid, attr)

        if (status > 0):
            if logger().DEBUG:
                logger().log_error(f'Reading variable (GET_EFIVAR) did not succeed: {status_dict.get(status, "UNKNOWN")} ({status:d})')
            data = b''
            guid = ''
            attr = 0
        else:
            data = buffer[base:base + new_size].tobytes()
            attr = struct.unpack('I', buffer[8:12])[0]
        return (off, buf, hdr, data, guid, attr)

    def kern_get_EFI_variable(self, name: str, guid: str) -> bytes:
        (_, _, _, data, guid, _) = self.kern_get_EFI_variable_full(name, guid)
        return data

    def kern_list_EFI_variables(self) -> Optional[Dict[str, List['EfiVariableType']]]:
        varlist = []
        off = 0
        hdr = 0
        attr = 0
        try:
            if os.path.isdir('/sys/firmware/efi/efivars'):
                varlist = os.listdir('/sys/firmware/efi/efivars')
            elif os.path.isdir('/sys/firmware/efi/vars'):
                varlist = os.listdir('/sys/firmware/efi/vars')
            else:
                return None
        except Exception:
            if logger().DEBUG:
                logger().log_error('Failed to read /sys/firmware/efi/[vars|efivars]. Folder does not exist')
            return None
        variables = dict()
        for v in varlist:
            name = v[:-37]
            guid = v[len(name) + 1:]
            if name and name is not None:
                variables[name] = []
                var = self.kern_get_EFI_variable_full(name, guid)
                (off, buf, hdr, data, guid, attr) = var
                variables[name].append(var)
        return variables

    def kern_set_EFI_variable(self, name: str, guid: str, value: bytes, attr: Optional[int] = None) -> int:
        attr = 0x7 if attr is None else attr
        status_dict = {
            0: 'EFI_SUCCESS',
            1: 'EFI_LOAD_ERROR',
            2: 'EFI_INVALID_PARAMETER',
            3: 'EFI_UNSUPPORTED',
            4: 'EFI_BAD_BUFFER_SIZE',
            5: 'EFI_BUFFER_TOO_SMALL',
            6: 'EFI_NOT_READY',
            7: 'EFI_DEVICE_ERROR',
            8: 'EFI_WRITE_PROTECTED',
            9: 'EFI_OUT_OF_RESOURCES',
            14: 'EFI_NOT_FOUND',
            26: 'EFI_SECURITY_VIOLATION'
        }

        header_size = 60  # 4*15
        namelen = len(name)
        if value:
            if isinstance(value, str):
                value = str.encode(value)
            datalen = len(value)
        else:
            datalen = 0
            value = struct.pack('B', 0x0)
        data_size = header_size + namelen + datalen
        guid0 = int(guid[:8], 16)
        guid1 = int(guid[9:13], 16)
        guid2 = int(guid[14:18], 16)
        guid3 = int(guid[19:21], 16)
        guid4 = int(guid[21:23], 16)
        guid5 = int(guid[24:26], 16)
        guid6 = int(guid[26:28], 16)
        guid7 = int(guid[28:30], 16)
        guid8 = int(guid[30:32], 16)
        guid9 = int(guid[32:34], 16)
        guid10 = int(guid[34:], 16)

        pack_formatting = f'15I{namelen}s{datalen}s'
        _guid = (guid0, guid1, guid2, guid3, guid4, guid5, guid6, guid7, guid8, guid9, guid10)
        in_buf = struct.pack(pack_formatting, data_size, *_guid, attr, namelen, datalen, name.encode('utf-8'), value)
        buffer = array.array('B', in_buf)
        self.ioctl(IOCTL_SET_EFIVAR, buffer)
        _, status = struct.unpack('2I', buffer[:8])

        if (status != 0):
            if logger().DEBUG:
                logger().log_error(f"Setting EFI (SET_EFIVAR) variable did not succeed: '{status_dict.get(status, 'UNKNOWN')}' ({status:d})")
        else:
            os.system('umount /sys/firmware/efi/efivars; mount -t efivarfs efivarfs /sys/firmware/efi/efivars')
        return status

    def hypercall(self, rcx: int, rdx: int, r8: int, r9: int, r10: int, r11: int, rax: int, rbx: int, rdi: int, rsi: int, xmm_buffer: int) -> int:
        in_buf = struct.pack(f'<11{self._pack}', rcx, rdx, r8, r9, r10, r11, rax, rbx, rdi, rsi, xmm_buffer)
        out_buf = self.ioctl(IOCTL_HYPERCALL, in_buf)
        return struct.unpack(f'<11{self._pack}', out_buf)[0]

    def send_sw_smi(self, cpu_thread_id: int, SMI_code_data: int, _rax: int, _rbx: int, _rcx: int, _rdx: int, _rsi: int, _rdi: int) -> Optional[Tuple[int, int, int, int, int, int, int]]:
        self.set_affinity(cpu_thread_id)
        in_buf = struct.pack(f'7{self._pack}', SMI_code_data, _rax, _rbx, _rcx, _rdx, _rsi, _rdi)
        out_buf = self.ioctl(IOCTL_SWSMI, in_buf)
        ret = struct.unpack(f'7{self._pack}', out_buf)
        return ret

    def send_sw_smi_timed(self, cpu_thread_id: int, SMI_code_data: int, _rax: int, _rbx: int, _rcx: int, _rdx: int, _rsi: int, _rdi: int) -> Optional[Tuple[int, int, int, int, int, int, int]]:
        self.set_affinity(cpu_thread_id)
        in_buf = struct.pack(f'8{self._pack}', SMI_code_data, _rax, _rbx, _rcx, _rdx, _rsi, _rdi, 0)
        out_buf = self.ioctl(IOCTL_SWSMI_TIMED, in_buf)
        ret = struct.unpack(f'8{self._pack}', out_buf)
        return ret

    def get_tool_info(self, tool_type: str) -> Tuple[Optional[str], str]:
        tool_name = _tools[tool_type] if tool_type in _tools else None
        tool_path = os.path.join(get_tools_path(), self.os_system.lower())
        return tool_name, tool_path

    def get_page_is_ram(self) -> Optional[bytes]:
        PROC_KALLSYMS = '/proc/kallsyms'
        symarr = chipsec.library.file.read_file(PROC_KALLSYMS).splitlines()
        for line in symarr:
            if b'page_is_ram' in line:
                return line.split(b' ')[0]
        return None

    def get_phys_mem_access_prot(self) -> Optional[bytes]:
        PROC_KALLSYMS = '/proc/kallsyms'
        symarr = chipsec.library.file.read_file(PROC_KALLSYMS).splitlines()
        for line in symarr:
            if b'phys_mem_access_prot' in line:
                return line.split(b' ')[0]
        return None

    def get_threads_count(self) -> int:
        import multiprocessing
        return multiprocessing.cpu_count()

    def retpoline_enabled(self):
        raise NotImplementedError('retpoline_enabled')


def get_helper():
    return LinuxHelper()



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

# -------------------------------------------------------------------------------
#
# CHIPSEC: Platform Hardware Security Assessment Framework
#
# -------------------------------------------------------------------------------

"""
On UEFI use the efi package functions
"""

import struct
import sys
import uuid
import os
import edk2   # Python 3.6.8 on UEFI

from typing import Dict, List, Optional, Tuple, TYPE_CHECKING
if TYPE_CHECKING:
    from ctypes import Array
    from chipsec.library.types import EfiVariableType
from chipsec.library.logger import logger
from chipsec.helper.oshelper import get_tools_path
from chipsec.helper.basehelper import Helper
from chipsec.library.exceptions import UnimplementedAPIError


_tools = {
}


class EfiHelper(Helper):

    def __init__(self):
        super(EfiHelper, self).__init__()
        self.name = "EfiHelper"
        if sys.platform.startswith('EFI'):
            self.os_system = sys.platform
            self.os_release = "0.0"
            self.os_version = "0.0"
            self.os_machine = "i386"
        else:
            import platform
            self.os_system = platform.system()
            self.os_release = platform.release()
            self.os_version = platform.version()
            self.os_machine = platform.machine()
            self.os_uname = platform.uname()

    def __del__(self):
        try:
            destroy()
        except NameError:
            pass

###############################################################################################
# Driver/service management functions
###############################################################################################

    def create(self) -> bool:
        logger().log_debug('[helper] UEFI Helper created')
        return True

    def start(self) -> bool:
        # The driver is part of the modified version of edk2.
        # It is always considered as loaded.
        self.driver_loaded = True
        logger().log_debug('[helper] UEFI Helper started/loaded')
        return True

    def stop(self) -> bool:
        logger().log_debug('[helper] UEFI Helper stopped/unloaded')
        return True

    def delete(self) -> bool:
        logger().log_debug('[helper] UEFI Helper deleted')
        return True

###############################################################################################
# Actual API functions to access HW resources
###############################################################################################

    #
    # Physical memory access
    #
    def split_address(self, pa: int) -> Tuple[int, int]:
        return (pa & 0xFFFFFFFF, (pa >> 32) & 0xFFFFFFFF)

    def read_phys_mem(self, phys_address: int, length: int) -> bytes:
        pa_lo, pa_hi = self.split_address(phys_address)
        return edk2.readmem(pa_lo, pa_hi, length)

    def write_phys_mem(self, phys_address: int, length: int, buf: bytes) -> int:
        pa_lo, pa_hi = self.split_address(phys_address)
        if type(buf) == bytearray:
            buf = bytes(buf)
        if 4 == length:
            dword_value = struct.unpack('I', buf)[0]
            res = edk2.writemem_dword(pa_lo, pa_hi, dword_value)
        else:
            res = edk2.writemem(pa_lo, pa_hi, buf)
        return res

    def alloc_phys_mem(self, length: int, max_pa: int) -> Tuple[int, int]:
        va = edk2.allocphysmem(length, max_pa)[0]
        (pa, _) = self.va2pa(va)
        return (va, pa)

    def va2pa(self, va: int) -> Tuple[int, int]:
        pa = va  # UEFI shell has identity mapping
        logger().log_debug(f'[helper] VA (0X{va:016X}) -> PA (0X{pa:016X})')
        return (pa, 0)

    def pa2va(self, pa: int) -> int:
        va = pa  # UEFI Shell has identity mapping
        logger().log_debug(f'[helper] PA (0X{pa:016X}) -> VA (0X{va:016X})')
        return va

    #
    # Memory-mapped I/O (MMIO) access
    #

    def map_io_space(self, physical_address: int, length: int, cache_type: int) -> int:
        return self.pa2va(physical_address)

    def read_mmio_reg(self, phys_address: int, size: int) -> int:
        phys_address_lo = phys_address & 0xFFFFFFFF
        phys_address_hi = (phys_address >> 32) & 0xFFFFFFFF
        out_buf = edk2.readmem(phys_address_lo, phys_address_hi, size)
        if size == 8:
            value = struct.unpack('=Q', out_buf[:size])[0]
        elif size == 4:
            value = struct.unpack('=I', out_buf[:size])[0]
        elif size == 2:
            value = struct.unpack('=H', out_buf[:size])[0]
        elif size == 1:
            value = struct.unpack('=B', out_buf[:size])[0]
        else:
            value = 0
        return value

    def write_mmio_reg(self, phys_address: int, size: int, value: int) -> int:
        phys_address_lo = phys_address & 0xFFFFFFFF
        phys_address_hi = (phys_address >> 32) & 0xFFFFFFFF
        if size == 4:
            ret = edk2.writemem_dword(phys_address_lo, phys_address_hi, value)
        else:
            buf = struct.pack(size * "B", value)
            ret = edk2.writemem(phys_address_lo, phys_address_hi, buf)
        return ret

    #
    # PCIe configuration access
    #

    def read_pci_reg(self, bus: int, device: int, function: int, address: int, size: int) -> int:
        if (1 == size):
            return (edk2.readpci(bus, device, function, address, size) & 0xFF)
        elif (2 == size):
            return (edk2.readpci(bus, device, function, address, size) & 0xFFFF)
        else:
            return edk2.readpci(bus, device, function, address, size)

    def write_pci_reg(self, bus: int, device: int, function: int, address: int, value: int, size: int) -> int:
        return edk2.writepci(bus, device, function, address, value, size)

    #
    # CPU I/O port access
    #

    def read_io_port(self, io_port: int, size: int) -> int:
        if (1 == size):
            return (edk2.readio(io_port, size) & 0xFF)
        elif (2 == size):
            return (edk2.readio(io_port, size) & 0xFFFF)
        else:
            return edk2.readio(io_port, size)

    def write_io_port(self, io_port: int, value: int, size: int) -> int:
        return edk2.writeio(io_port, size, value)

    #
    # SMI events
    #

    def send_sw_smi(self, cpu_thread_id: int, SMI_code_data: int, _rax: int, _rbx: int, _rcx: int, _rdx: int, _rsi: int, _rdi: int) -> None:
        return edk2.swsmi(SMI_code_data, _rax, _rbx, _rcx, _rdx, _rsi, _rdi)

    #
    # CPU related API
    #

    def read_msr(self, cpu_thread_id: int, msr_addr: int) -> Tuple[int, int]:
        (eax, edx) = edk2.rdmsr(msr_addr)
        eax = eax % 2**32
        edx = edx % 2**32
        return (eax, edx)

    def write_msr(self, cpu_thread_id: int, msr_addr: int, eax: int, edx: int) -> int:
        return edk2.wrmsr(msr_addr, eax, edx)

    def read_cr(self, cpu_thread_id: int, cr_number: int) -> int:
        raise UnimplementedAPIError('read_cr')

    def write_cr(self, cpu_thread_id: int, cr_number: int, value: int) -> int:
        raise UnimplementedAPIError('write_cr')

    def load_ucode_update(self, cpu_thread_id: int, ucode_update_buf: int) -> bool:
        raise UnimplementedAPIError('load_ucode_update')

    def get_threads_count(self) -> int:
        return 1

    def cpuid(self, eax: int, ecx: int) -> Tuple[int, int, int, int]:
        (reax, rebx, recx, redx) = edk2.cpuid(eax, ecx)
        return (reax, rebx, recx, redx)

    def get_descriptor_table(self, cpu_thread_id: int, desc_table_code: int)-> None:
        raise UnimplementedAPIError('get_descriptor_table')

    #
    # File system
    #

    def get_tool_info(self, tool_type: str) -> Tuple[str, str]:
        tool_name = _tools[tool_type] if tool_type in _tools else ''
        tool_path = os.path.join(get_tools_path(), self.os_system.lower())
        return (tool_name, tool_path)

    #
    # EFI Variable API
    #

    def EFI_supported(self) -> bool:
        return True

    def get_EFI_variable_full(self, name: str, guidstr: str) -> Tuple[int, Optional[bytes], int]:

        size = 100
        (Status, Attributes, newdata, DataSize) = edk2.GetVariable(name, guidstr, size)

        if Status == 5:
            size = DataSize + 1
            (Status, Attributes, newdata, DataSize) = edk2.GetVariable(name, guidstr, size)

        return (Status, newdata, Attributes)

    def get_EFI_variable(self, name: str, guidstr: str) -> Optional[bytes]:
        (_, data, _) = self.get_EFI_variable_full(name, guidstr)
        return data

    def set_EFI_variable(self, name: str, guidstr: str, buffer: bytes, buffer_size: Optional[int] = None, attrs: Optional[int] = 0x7) -> int:

        if buffer_size is None:
            buffer_size = len(buffer)
        if attrs is None:
            attrs = 0x07
            if logger().VERBOSE:
                logger().log_important(f'Setting attributes to: {attrs:04X}')
        elif isinstance(attrs, bytes):
            attrs = struct.unpack("L", attrs)[0]

        (Status, buffer_size, guidstr) = edk2.SetVariable(name, guidstr, int(attrs), buffer, buffer_size)

        return Status

    def delete_EFI_variable(self, name: str, guid: str) -> int:
        return self.set_EFI_variable(name, guid, bytes(4), 0, 0)

    def list_EFI_variables(self) -> Optional[Dict[str, List['EfiVariableType']]]:

        off = 0
        buf = b''
        hdr = 0
        attr = 0
        var_list = list()
        variables = dict()

        status_dict = {0: "EFI_SUCCESS", 1: "EFI_LOAD_ERROR", 2: "EFI_INVALID_PARAMETER", 3: "EFI_UNSUPPORTED", 4: "EFI_BAD_BUFFER_SIZE", 5: "EFI_BUFFER_TOO_SMALL",
                       6: "EFI_NOT_READY", 7: "EFI_DEVICE_ERROR", 8: "EFI_WRITE_PROTECTED", 9: "EFI_OUT_OF_RESOURCES", 14: "EFI_NOT_FOUND", 26: "EFI_SECURITY_VIOLATION"}

        namestr = ''
        size = 200
        guidstr = str(uuid.uuid4())

        search_complete = False
        while not search_complete:
            namestr += '\x00'
            name = namestr.encode('utf-16-le')
            guid = uuid.UUID(guidstr).bytes_le
            (status, namestr, size, guidstr) = edk2.GetNextVariableName(size, name, guid)

            if status == 5:
                logger().log_debug(f'[helper] EFI Variable name size was too small increasing to {size:d}')
                (status, namestr, size, guidstr) = edk2.GetNextVariableName(size, name, guid)

            logger().log_debug(f'[helper] Returned {name}. Status is {status_dict[status]}')

            if status:
                search_complete = True
            else:
                if (namestr, guidstr) in var_list:
                    continue
                else:
                    var_list.append((namestr, guidstr))

                logger().log_debug(f"[helper] Found variable '{name}' - [{guidstr}]")

        for (name, guidstr) in var_list:
            (status, data, attr) = self.get_EFI_variable_full(name, guidstr)

            if status:
                logger().log_verbose(f'[helper] Error reading variable {name}.  Status = {status:d} - {status_dict[status]}')

            var_data = (off, buf, hdr, data, guidstr, attr)

            if name in variables:
                logger().log_verbose(f'[helper] Duplicate variable name {name} - {guidstr}')
                continue
            else:
                variables[name] = []

            if data != '' or guidstr != '' or attr != 0:
                variables[name].append(var_data)

        return variables

    #
    # ACPI tables access
    #

    #
    # IOSF Message Bus access
    #

    def msgbus_send_read_message(self, mcr: int, mcrx: int) -> None:
        raise UnimplementedAPIError('msgbus_send_read_message')

    def msgbus_send_write_message(self, mcr: int, mcrx: int, mdr: int) -> None:
        raise UnimplementedAPIError('msgbus_send_write_message')

    def msgbus_send_message(self, mcr: int, mcrx: int, mdr: Optional[int] = None) -> None:
        raise UnimplementedAPIError('msgbus_send_message')

    def set_affinity(self, value: int) -> None:
        raise UnimplementedAPIError('set_affinity')

    def free_phys_mem(self, physical_address):
        raise UnimplementedAPIError('free_phys_mem')

    def get_ACPI_table(self, table_name: str) -> Optional['Array']:
        raise UnimplementedAPIError('get_ACPI_table')

    def enum_ACPI_tables(self) -> Optional['Array']:
        raise UnimplementedAPIError('enum_ACPI_table')

    def get_affinity(self):
        raise UnimplementedAPIError('get_affinity')

    def hypercall(self, rcx, rdx, r8, r9, r10, r11, rax, rbx, rdi, rsi, xmm_buffer):
        raise UnimplementedAPIError('hypercall')

    def retpoline_enabled(self) -> bool:
        return False


def get_helper() -> EfiHelper:
    return EfiHelper()



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


import os
import sys

print('Python', sys.version)

WIN_DRIVER_INSTALL_PATH = 'chipsec/helper/windows'
VERSION_FILE = 'VERSION'

build_dir = os.getcwd()
root_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), os.path.pardir))
bin_dir = os.path.join(root_dir, 'bin')
tool_dir = root_dir
cfg_dir = os.path.join(tool_dir, 'chipsec', 'cfg')
version_file = os.path.join(root_dir, 'chipsec', VERSION_FILE)

win_7_amd64 = os.path.join(bin_dir, 'windows-amd64')

print(os.getcwd())
os.chdir(tool_dir)
sys.path.append(tool_dir)
print(os.getcwd())

data_files = [
    (WIN_DRIVER_INSTALL_PATH + '/windows_amd64', ['chipsec/helper/windows/windows_amd64/chipsec_hlpr.sys']),
    ('chipsec/modules/tools/uefi', ['chipsec/modules/tools/uefi/blockedlist.json']),
    ('chipsec/modules/tools/secureboot', [
        'chipsec/modules/tools/secureboot/Shell.efi',
        'chipsec/modules/tools/secureboot/te.cfg'
    ]),
]

for current, dirs, files in os.walk(cfg_dir):
    for file in files:
        if file.endswith('.xml'):
            tail = current.replace(cfg_dir, '').replace('\\', '/')
            xf = f'chipsec/cfg{tail}', [f'chipsec/cfg{tail}/{file}']
            data_files.append(xf)

version = ''
if os.path.exists(version_file):
    data_files.append(('.', [version_file]))
    with open(version_file, 'r') as verFile:
        version = verFile.read()
print(f'VERSION: {version}')

mypackages = []
for current, dirs, files in os.walk(tool_dir):
    if current.startswith(os.path.join(tool_dir, 'build')):
        continue
    for file in files:
        if file == '__init__.py':
            pkg = current.replace(tool_dir + os.path.sep, '')
            pkg = pkg.replace(os.path.sep, '.')
            mypackages.append(pkg)
            print(pkg)

if not os.path.exists(win_7_amd64):
    os.makedirs(win_7_amd64)

from setuptools import setup

includes = []

setup(
    name='chipsec',
    description='CHIPSEC: Platform Security Assessment Framework',
    version=version,
    console=['chipsec_main.py', 'chipsec_util.py'],
    data_files=data_files,
    options={
        'build': {'build_base': build_dir}
    }
)



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2016, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


"""
Hyper-V specific defines
"""

import re

msrs = {
    0x40000000: 'HV_X64_MSR_GUEST_OS_ID',
    0x40000001: 'HV_X64_MSR_HYPERCALL',
    0x40000002: 'HV_X64_MSR_VP_INDEX',
    0x40000003: 'HV_X64_MSR_RESET',
    0x40000010: 'HV_X64_MSR_VP_RUNTIME',
    0x40000020: 'HV_X64_MSR_TIME_REF_COUNT',
    0x40000021: 'HV_X64_MSR_REFERENCE_TSC',
    0x40000022: 'HV_X64_MSR_TSC_FREQUENCY',
    0x40000023: 'HV_X64_MSR_APIC_FREQUENCY',
    0x40000070: 'HV_X64_MSR_EOI',
    0x40000071: 'HV_X64_MSR_ICR',
    0x40000072: 'HV_X64_MSR_TPR',
    0x40000073: 'HV_X64_MSR_APIC_ASSIST_PAGE',
    0x40000080: 'HV_X64_MSR_SCONTROL',
    0x40000081: 'HV_X64_MSR_SVERSION',
    0x40000082: 'HV_X64_MSR_SIEFP',
    0x40000083: 'HV_X64_MSR_SIMP',
    0x40000084: 'HV_X64_MSR_EOM',
    0x40000090: 'HV_X64_MSR_SINT0',
    0x40000091: 'HV_X64_MSR_SINT1',
    0x40000092: 'HV_X64_MSR_SINT2',
    0x40000093: 'HV_X64_MSR_SINT3',
    0x40000094: 'HV_X64_MSR_SINT4',
    0x40000095: 'HV_X64_MSR_SINT5',
    0x40000096: 'HV_X64_MSR_SINT6',
    0x40000097: 'HV_X64_MSR_SINT7',
    0x40000098: 'HV_X64_MSR_SINT8',
    0x40000099: 'HV_X64_MSR_SINT9',
    0x4000009A: 'HV_X64_MSR_SINT10',
    0x4000009B: 'HV_X64_MSR_SINT11',
    0x4000009C: 'HV_X64_MSR_SINT12',
    0x4000009D: 'HV_X64_MSR_SINT13',
    0x4000009E: 'HV_X64_MSR_SINT14',
    0x4000009F: 'HV_X64_MSR_SINT15',
    0x400000B0: 'HV_X64_MSR_STIMER0_CONFIG',
    0x400000B1: 'HV_X64_MSR_STIMER0_COUNT',
    0x400000B2: 'HV_X64_MSR_STIMER1_CONFIG',
    0x400000B3: 'HV_X64_MSR_STIMER1_COUNT',
    0x400000B4: 'HV_X64_MSR_STIMER2_CONFIG',
    0x400000B5: 'HV_X64_MSR_STIMER2_COUNT',
    0x400000B6: 'HV_X64_MSR_STIMER3_CONFIG',
    0x400000B7: 'HV_X64_MSR_STIMER3_COUNT',
    0x400000C1: 'HV_X64_MSR_POWER_STATE_TRIGGER_C1',
    0x400000C2: 'HV_X64_MSR_POWER_STATE_TRIGGER_C2',
    0x400000C3: 'HV_X64_MSR_POWER_STATE_TRIGGER_C3',
    0x400000D1: 'HV_X64_MSR_POWER_STATE_CONFIG_C1',
    0x400000D2: 'HV_X64_MSR_POWER_STATE_CONFIG_C2',
    0x400000D3: 'HV_X64_MSR_POWER_STATE_CONFIG_C3',
    0x400000E0: 'HV_X64_MSR_STATS_PARTITION_RETAIL_PAGE',
    0x400000E1: 'HV_X64_MSR_STATS_PARTITION_INTERNAL_PAGE',
    0x400000E2: 'HV_X64_MSR_STATS_VP_RETAIL_PAGE',
    0x400000E3: 'HV_X64_MSR_STATS_VP_INTERNAL_PAGE',
    0x400000F0: 'HV_X64_MSR_GUEST_IDLE',
    0x400000F1: 'HV_X64_MSR_SYNTH_DEBUG_CONTROL',
    0x400000F2: 'HV_X64_MSR_SYNTH_DEBUG_STATUS',
    0x400000F3: 'HV_X64_MSR_SYNTH_DEBUG_SEND_BUFFER',
    0x400000F4: 'HV_X64_MSR_SYNTH_DEBUG_RECEIVE_BUFFER',
    0x400000F5: 'HV_X64_MSR_SYNTH_DEBUG_PENDING_BUFFER',
    0x40000100: 'HV_X64_MSR_CRASH_P0',
    0x40000101: 'HV_X64_MSR_CRASH_P1',
    0x40000102: 'HV_X64_MSR_CRASH_P2',
    0x40000103: 'HV_X64_MSR_CRASH_P3',
    0x40000104: 'HV_X64_MSR_CRASH_P4',
    0x40000105: 'HV_X64_MSR_CRASH_CTL'
}


def get_msr_name(code, defvalue=''):
    return msrs[code] if code in msrs else defvalue


hypercall_status_codes = {
    0x0000: 'HV_STATUS_SUCCESS',
    0x0001: 'HV_RESERVED_01H',
    0x0002: 'HV_STATUS_INVALID_HYPERCALL_CODE',
    0x0003: 'HV_STATUS_INVALID_HYPERCALL_INPUT',
    0x0004: 'HV_STATUS_INVALID_ALIGNMENT',
    0x0005: 'HV_STATUS_INVALID_PARAMETER',
    0x0006: 'HV_STATUS_ACCESS_DENIED',
    0x0007: 'HV_STATUS_INVALID_PARTITION_STATE',
    0x0008: 'HV_STATUS_OPERATION_DENIED',
    0x0009: 'HV_STATUS_UNKNOWN_PROPERTY',
    0x000A: 'HV_STATUS_PROPERTY_VALUE_OUT_OF_RANGE',
    0x000B: 'HV_STATUS_INSUFFICIENT_MEMORY',
    0x000C: 'HV_STATUS_PARTITION_TOO_DEEP',
    0x000D: 'HV_STATUS_INVALID_PARTITION_ID',
    0x000E: 'HV_STATUS_INVALID_VP_INDEX',
    0x000F: 'HV_RESERVED_0FH',
    0x0010: 'HV_RESERVED_10H',
    0x0011: 'HV_STATUS_INVALID_PORT_ID',
    0x0012: 'HV_STATUS_INVALID_CONNECTION_ID',
    0x0013: 'HV_STATUS_INSUFFICIENT_BUFFERS',
    0x0014: 'HV_STATUS_NOT_ACKNOWLEDGED',
    0x0015: 'HV_RESERVED_15H',
    0x0016: 'HV_STATUS_ACKNOWLEDGED',
    0x0017: 'HV_STATUS_INVALID_SAVE_RESTORE_STATE',
    0x0018: 'HV_STATUS_INVALID_SYNIC_STATE',
    0x0019: 'HV_STATUS_OBJECT_IN_USE',
    0x001A: 'HV_STATUS_INVALID_PROXIMITY_DOMAIN_INFO',
    0x001B: 'HV_STATUS_NO_DATA',
    0x001C: 'HV_STATUS_INACTIVE',
    0x001D: 'HV_STATUS_NO_RESOURCES',
    0x001E: 'HV_STATUS_FEATURE_UNAVAILABLE',
    0x001F: 'HV_STATUS_PARTIAL_PACKET',
    0x0020: 'HV_STATUS_PROCESSOR_FEATURE_SSE3_NOT_SUPPORTED',
    0x0021: 'HV_STATUS_PROCESSOR_FEATURE_LAHFSAHF_NOT_SUPPORTED',
    0x0022: 'HV_STATUS_PROCESSOR_FEATURE_SSSE3_NOT_SUPPORTED',
    0x0023: 'HV_STATUS_PROCESSOR_FEATURE_SSE4_1_NOT_SUPPORTED',
    0x0024: 'HV_STATUS_PROCESSOR_FEATURE_SSE4_2_NOT_SUPPORTED',
    0x0025: 'HV_STATUS_PROCESSOR_FEATURE_SSE4A_NOT_SUPPORTED',
    0x0026: 'HV_STATUS_PROCESSOR_FEATURE_XOP_NOT_SUPPORTED',
    0x0027: 'HV_STATUS_PROCESSOR_FEATURE_POPCNT_NOT_SUPPORTED',
    0x0028: 'HV_STATUS_PROCESSOR_FEATURE_CMPXCHG16B_NOT_SUPPORTED',
    0x0029: 'HV_STATUS_PROCESSOR_FEATURE_ALTMOVCR8_NOT_SUPPORTED',
    0x002A: 'HV_STATUS_PROCESSOR_FEATURE_LZCNT_NOT_SUPPORTED',
    0x002B: 'HV_STATUS_PROCESSOR_FEATURE_MISALIGNED_SSE_NOT_SUPPORTED',
    0x002C: 'HV_STATUS_PROCESSOR_FEATURE_MMX_EXT_NOT_SUPPORTED',
    0x002D: 'HV_STATUS_PROCESSOR_FEATURE_3DNOW_NOT_SUPPORTED',
    0x002E: 'HV_STATUS_PROCESSOR_FEATURE_EXTENDED_3DNOW_NOT_SUPPORTED',
    0x002F: 'HV_STATUS_PROCESSOR_FEATURE_PAGE_1GB_NOT_SUPPORTED',
    0x0030: 'HV_STATUS_PROCESSOR_CACHE_LINE_FLUSH_SIZE_INCOMPATIBLE',
    0x0031: 'HV_STATUS_PROCESSOR_FEATURE_XSAVE_NOT_SUPPORTED',
    0x0032: 'HV_STATUS_PROCESSOR_FEATURE_XSAVEOPT_NOT_SUPPORTED',
    0x0033: 'HV_STATUS_INSUFFICIENT_BUFFER',
    0x0034: 'HV_STATUS_PROCESSOR_FEATURE_XSAVE_AVX_NOT_SUPPORTED',
    0x0035: 'HV_STATUS_PROCESSOR_FEATURE_XSAVE_FEATURE_NOT_SUPPORTED',
    0x0036: 'HV_STATUS_PROCESSOR_XSAVE_SAVE_AREA_INCOMPATIBLE',
    0x0037: 'HV_STATUS_INCOMPATIBLE_PROCESSOR',
    0x0038: 'HV_STATUS_INSUFFICIENT_DEVICE_DOMAINS',
    0x0039: 'HV_STATUS_PROCESSOR_FEATURE_AES_NOT_SUPPORTED',
    0x003A: 'HV_STATUS_PROCESSOR_FEATURE_PCLMULQDQ_NOT_SUPPORTED',
    0x003B: 'HV_STATUS_PROCESSOR_FEATURE_INCOMPATIBLE_XSAVE_FEATURES',
    0x003C: 'HV_STATUS_CPUID_FEATURE_VALIDATION_ERROR',
    0x003D: 'HV_STATUS_CPUID_XSAVE_FEATURE_VALIDATION_ERROR',
    0x003E: 'HV_STATUS_PROCESSOR_STARTUP_TIMEOUT',
    0x003F: 'HV_STATUS_SMX_ENABLED',
    0x0040: 'HV_STATUS_PROCESSOR_FEATURE_PCID_NOT_SUPPORTED',
    0x0041: 'HV_STATUS_INVALID_LP_INDEX',
    0x0042: 'HV_STATUS_FEATURE_FMA4_NOT_SUPPORTED',
    0x0043: 'HV_STATUS_FEATURE_F16C_NOT_SUPPORTED',
    0x0044: 'HV_STATUS_PROCESSOR_FEATURE_RDRAND_NOT_SUPPORTED',
    0x0045: 'HV_STATUS_PROCESSOR_FEATURE_RDWRFSGS_NOT_SUPPORTED',
    0x0046: 'HV_STATUS_PROCESSOR_FEATURE_SMEP_NOT_SUPPORTED',
    0x0047: 'HV_STATUS_PROCESSOR_FEATURE_ENHANCED_FAST_STRING_NOT_SUPPORTED',
    0x0048: 'HV_STATUS_PROCESSOR_FEATURE_MOVBE_NOT_SUPPORTED',
    0x0049: 'HV_STATUS_PROCESSOR_FEATURE_BMI1_NOT_SUPPORTED',
    0x004A: 'HV_STATUS_PROCESSOR_FEATURE_BMI2_NOT_SUPPORTED',
    0x004B: 'HV_STATUS_PROCESSOR_FEATURE_HLE_NOT_SUPPORTED',
    0x004C: 'HV_STATUS_PROCESSOR_FEATURE_RTM_NOT_SUPPORTED',
    0x004D: 'HV_STATUS_PROCESSOR_FEATURE_XSAVE_FMA_NOT_SUPPORTED',
    0x004E: 'HV_STATUS_PROCESSOR_FEATURE_XSAVE_AVX2_NOT_SUPPORTED',
    0x004F: 'HV_STATUS_PROCESSOR_FEATURE_NPIEP1_NOT_SUPPORTED'
}


def get_hypercall_status(code, defvalue=''):
    return hypercall_status_codes[code] if code in hypercall_status_codes else defvalue


hypercall_names = {
    0x0001: 'HvSwitchVirtualAddressSpace',
    0x0002: 'HvFlushVirtualAddressSpace',
    0x0003: 'HvFlushVirtualAddressList',
    0x0004: 'HvGetLogicalProcessorRunTime',
    0x0008: 'HvNotifyLongSpinWait',
    0x0009: 'HvParkedVirtualProcessors',
    0x0040: 'HvCreatePartition',
    0x0041: 'HvInitializePartition',
    0x0042: 'HvFinalizePartition',
    0x0043: 'HvDeletePartition',
    0x0044: 'HvGetPartitionProperty',
    0x0045: 'HvSetPartitionProperty',
    0x0046: 'HvGetPartitionId',
    0x0047: 'HvGetNextChildPartition',
    0x0048: 'HvDepositMemory',
    0x0049: 'HvWithdrawMemory',
    0x004A: 'HvGetMemoryBalance',
    0x004B: 'HvMapGpaPages',
    0x004C: 'HvUnmapGpaPages',
    0x004D: 'HvInstallIntercept',
    0x004E: 'HvCreateVp',
    0x004F: 'HvDeleteVp',
    0x0050: 'HvGetVpRegisters',
    0x0051: 'HvSetVpRegisters',
    0x0052: 'HvTranslateVirtualAddress',
    0x0053: 'HvReadGpa',
    0x0054: 'HvWriteGpa',
    0x0055: 'HvAssertVirtualInterrupt',
    0x0056: 'HvClearVirtualInterrupt',
    0x0057: 'HvCreatePort',
    0x0058: 'HvDeletePort',
    0x0059: 'HvConnectPort',
    0x005A: 'HvGetPortProperty',
    0x005B: 'HvDisconnectPort',
    0x005C: 'HvPostMessage',
    0x005D: 'HvSignalEvent',
    0x005E: 'HvSavePartitionState',
    0x005F: 'HvRestorePartitionState',
    0x0060: 'HvInitializeEventLogBufferGroup',
    0x0061: 'HvFinalizeEventLogBufferGroup',
    0x0062: 'HvCreateEventLogBuffer',
    0x0063: 'HvDeleteEventLogBuffer',
    0x0064: 'HvMapEventLogBuffer',
    0x0065: 'HvUnmapEventLogBuffer',
    0x0066: 'HvSetEventLogGroupSources',
    0x0067: 'HvReleaseEventLogBuffer',
    0x0068: 'HvFlushEventLogBuffer',
    0x0069: 'HvPostDebugData',
    0x006A: 'HvRetrieveDebugData',
    0x006B: 'HvResetDebugSession',
    0x006C: 'HvMapStatsPage',
    0x006D: 'HvUnmapStatsPage',
    0x006E: 'HvCallMapSparseGpaPages',
    0x006F: 'HvCallSetSystemProperty',
    0x0070: 'HvCallSetPortProperty',
    0x0076: 'HvCallAddLogicalProcessor',
    0x0077: 'HvCallRemoveLogicalProcessor',
    0x0078: 'HvCallQueryNumaDistance',
    0x0079: 'HvCallSetLogicalProcessorProperty',
    0x007A: 'HvCallGetLogicalProcessorProperty',
    0x007B: 'HvCallGetSystemProperty',
    0x007C: 'HvCallMapDeviceInterrupt',
    0x007D: 'HvCallUnmapDeviceInterrupt',
    0x007E: 'HvCallCreateDeviceDomain',
    0x007F: 'HvCallDeleteDeviceDomain',
    0x0080: 'HvCallMapDevicePages',
    0x0081: 'HvCallUnmapDevicePages',
    0x0082: 'HvCallAttachDevice',
    0x0083: 'HvCallDetachDevice',
    0x0084: 'HvCallEnterSleepState',
    0x0085: 'HvCallPrepareForSleep',
    0x0086: 'HvCallPrepareForHibernate',
    0x0087: 'HvCallNotifyPartitionEvent',
    0x0088: 'HvCallGetLogicalProcessorRegisters',
    0x0089: 'HvCallSetLogicalProcessorRegisters',
    0x008A: 'HvCallQueryAssociatedLpsforMca',
    0x008B: 'HvCallNotifyRingEmpty',
    0x008C: 'HvCallInjectSyntheticMachineCheck',
    0x008D: 'HvCallScrubPartition',
    0x008E: 'HvCallCollectLivedump',
    0x008F: 'HvCallDisableHypervisor',
    0x0090: 'HvCallModifySparseGpaPages',
    0x0091: 'HvCallRegisterInterceptResult',
    0x0092: 'HvCallUnregisterInterceptResult'
}


def get_hypercall_name(code, defvalue=''):
    return hypercall_names[code] if code in hypercall_names else defvalue


hv_porttype = {
    0x0001: 'HvPortTypeMessage',
    0x0002: 'HvPortTypeEvent',
    0x0003: 'HvPortTypeMonitor'
}

GOOD_PARAMS_STATUSES = ['HV_STATUS_SUCCESS', 'HV_STATUS_ACCESS_DENIED']

cpuid_desc = {0x40000003: {}}

cpuid_desc[0x40000003]['EAX'] = {
    0: 'VP Runtime (HV_X64_MSR_VP_RUNTIME) available',
    1: 'Partition Reference Counter (HV_X64_MSR_TIME_REF_COUNT) available',
    2: 'Basic SynIC MSRs (HV_X64_MSR_SCONTROL..HV_X64_MSR_EOM and HV_X64_MSR_SINTxx) available',
    3: 'Synthetic Timer MSRs (HV_X64_MSR_STIMER0_CONFIG..HV_X64_MSR_STIMER3_COUNT) available',
    4: 'APIC access MSRs (HV_X64_MSR_EOI, HV_X64_MSR_ICR and HV_X64_MSR_TPR) are available',
    5: 'Hypercall MSRs (HV_X64_MSR_GUEST_OS_ID and HV_X64_MSR_HYPERCALL) available',
    6: 'Access virtual processor index MSR (HV_X64_MSR_VP_INDEX) available',
    7: 'Virtual system reset MSR (HV_X64_MSR_RESET) is available',
    8: 'Access statistics pages MSRs available',
    9: 'Partition Reference TSC MSR (HV_X64_MSR_REFERENCE_TSC) available',
    10: 'Virtual Guest Idle State MSR (HV_X64_MSR_GUEST_IDLE) available',
    11: 'Timer frequency MSRs (HV_X64_MSR_TSC_FREQUENCY and HV_X64_MSR_APIC_FREQUENCY) available',
    12: 'Debug MSRs (HV_X64_MSR_SYNTH_DEBUG_*) available'
}

cpuid_desc[0x40000003]['EBX'] = {
    0: 'CreatePartitions',
    1: 'AccessPartitionId',
    2: 'AccessMemoryPool',
    3: 'AdjustMessageBuffers',
    4: 'PostMessages',
    5: 'SignalEvents',
    6: 'CreatePort',
    7: 'ConnectPort',
    8: 'AccessStats',
    9: 'RsvdZ',
    10: 'RsvdZ',
    11: 'Debugging',
    12: 'CpuManagement',
    13: 'ConfigureProfiler',
    14: 'EnableExpandedStackwalking'
}

cpuid_desc[0x40000003]['ECX'] = {
    0: 'Maximum Processor Power State (bit 0)',
    1: 'Maximum Processor Power State (bit 1)',
    2: 'Maximum Processor Power State (bit 2)',
    4: 'HPET is required to enter C3'
}

cpuid_desc[0x40000003]['EDX'] = {
    0: 'The MWAIT instruction is available',
    1: 'Guest debugging support is available',
    2: 'Performance Monitor support is available',
    3: 'Support for physical CPU dynamic partitioning events is available',
    4: 'Support for passing hypercall input parameter block via XMM registers is available',
    5: 'Support for a virtual guest idle state is available',
    6: 'Support for hypervisor sleep state is available',
    7: 'Support for querying NUMA distances is available',
    8: 'Support for determining timer frequencies is available',
    9: 'Support for injecting synthetic machine checks is available',
    10: 'Support for guest crash MSRs is available',
    11: 'Support for debug MSRs is available',
    12: 'Npiep1Available',
    13: 'DiableHypervisorAvailable'
}

### HYPERV DEFINES #############################################################

# VMBUS version is 32 bit entity broken up into two 16 bit quantities: major_number . minor_number
VERSION_WS2008 = ((0 << 16) | (13))  # 0 . 13 (Windows Server 2008)
VERSION_WIN7 = ((1 << 16) | (1))   # 1 . 1  (Windows 7)
VERSION_WIN8 = ((2 << 16) | (4))   # 2 . 4  (Windows 8)
VERSION_WIN8_1 = ((3 << 16) | (0))   # 3 . 0  (Windows 8 R2)

vmbus_versions = {
    VERSION_WS2008: 'Windows Server 2008',
    VERSION_WIN7: 'Windows 7',
    VERSION_WIN8: 'Windows 8',
    VERSION_WIN8_1: 'Windows 8 R2'
}

VMBUS_MESSAGE_CONNECTION_ID = 1
VMBUS_MESSAGE_PORT_ID = 1
VMBUS_EVENT_CONNECTION_ID = 2
VMBUS_EVENT_PORT_ID = 2
VMBUS_MONITOR_CONNECTION_ID = 3
VMBUS_MONITOR_PORT_ID = 3
VMBUS_MESSAGE_SINT = 2

### HV MESSAGE DEFINES #########################################################

# Define hypervisor message types
hv_message_type = {
    0x00000000: 'HVMSG_NONE',
    # Memory access messages
    0x80000000: 'HVMSG_UNMAPPED_GPA',
    0x80000001: 'HVMSG_GPA_INTERCEPT',
    # Timer notification messages
    0x80000010: 'HVMSG_TIMER_EXPIRED',
    # Error messages
    0x80000020: 'HVMSG_INVALID_VP_REGISTER_VALUE',
    0x80000021: 'HVMSG_UNRECOVERABLE_EXCEPTION',
    0x80000022: 'HVMSG_UNSUPPORTED_FEATURE',
    # Trace buffer complete messages
    0x80000040: 'HVMSG_EVENTLOG_BUFFERCOMPLETE',
    # Platform-specific processor intercept messages
    0x80010000: 'HVMSG_X64_IOPORT_INTERCEPT',
    0x80010001: 'HVMSG_X64_MSR_INTERCEPT',
    0x80010002: 'HVMSG_X64_CPUID_INTERCEPT',
    0x80010003: 'HVMSG_X64_EXCEPTION_INTERCEPT',
    0x80010004: 'HVMSG_X64_APIC_EOI',
    0x80010005: 'HVMSG_X64_LEGACY_FP_ERROR'
}

### MSG CHANNEL DEFINES ########################################################

CHANNELMSG_INVALID = 0
CHANNELMSG_OFFERCHANNEL = 1
CHANNELMSG_RESCIND_CHANNELOFFER = 2
CHANNELMSG_REQUESTOFFERS = 3
CHANNELMSG_ALLOFFERS_DELIVERED = 4
CHANNELMSG_OPENCHANNEL = 5
CHANNELMSG_OPENCHANNEL_RESULT = 6
CHANNELMSG_CLOSECHANNEL = 7
CHANNELMSG_GPADL_HEADER = 8
CHANNELMSG_GPADL_BODY = 9
CHANNELMSG_GPADL_CREATED = 10
CHANNELMSG_GPADL_TEARDOWN = 11
CHANNELMSG_GPADL_TORNDOWN = 12
CHANNELMSG_RELID_RELEASED = 13
CHANNELMSG_INITIATE_CONTACT = 14
CHANNELMSG_VERSION_RESPONSE = 15
CHANNELMSG_UNLOAD = 16

vmbus_channel_message_type = {
    0: 'ChannelMessageInvalid',
    1: 'ChannelMessageOfferChannel',
    2: 'ChannelMessageRescindChannelOffer',
    3: 'ChannelMessageRequestOffers',
    4: 'ChannelMessageAllOffersDelivered',
    5: 'ChannelMessageOpenChannel',
    6: 'ChannelMessageOpenChannelResult',
    7: 'ChannelMessageCloseChannel',
    8: 'ChannelMessageGpadlHeader',
    9: 'ChannelMessageGpadlBody',
    10: 'ChannelMessageGpadlCreated',
    11: 'ChannelMessageGpadlTeardown',
    12: 'ChannelMessageGpadlTorndown',
    13: 'ChannelMessageRelIdReleased',
    14: 'ChannelMessageInitiateContact',
    15: 'ChannelMessageVersionResponse',
    16: 'ChannelMessageUnload'
}

channel_flags = {
    0x00: 'VMBUS_CHANNEL_ENUMERATE_DEVICE_INTERFACE',
    0x01: 'VMBUS_CHANNEL_SERVER_SUPPORTS_TRANSFER_PAGES',
    0x02: 'VMBUS_CHANNEL_SERVER_SUPPORTS_GPADLS',
    0x04: 'VMBUS_CHANNEL_NAMED_PIPE_MODE',
    0x08: 'VMBUS_CHANNEL_LOOPBACK_OFFER',
    0x09: 'VMBUS_CHANNEL_PARENT_OFFER',
    0x0a: 'VMBUS_CHANNEL_REQUEST_MONITORED_NOTIFICATION'
}

# GUID definitions of various offer types - services offered to the guest
HV_NIC_GUID = '{f8615163-df3e-46c5-913f-f2d2f965ed0e}'
HV_IDE_GUID = '{32412632-86cb-44a2-9b5c-50d1417354f5}'
HV_SCSI_GUID = '{ba6163d9-04a1-4d29-b605-72e2ffb1dc7f}'
HV_SHUTDOWN_GUID = '{0e0b6031-5213-4934-818b-38d90ced39db}'
HV_TS_GUID = '{9527e630-d0ae-497b-adce-e80ab0175caf}'
HV_HEART_BEAT_GUID = '{57164f39-9115-4e78-ab55-382f3bd5422d}'
HV_KVP_GUID = '{a9a0f4e7-5a45-4d96-b827-8a841e8c03e6}'
HV_DM_GUID = '{525074dc-8985-46e2-8057-a307dc18a502}'
HV_MOUSE_GUID = '{cfa8b69e-5b4a-4cc0-b98b-8ba1a1f3f95a}'
HV_VSS_GUID = '{35fa2e29-ea23-4236-ae96-3a6ebacba440}'
HV_SYNTHVID_GUID = '{da0a7802-e377-4aac-8e77-0558eb1073f8}'
HV_SYNTHFC_GUID = '{2f9bcc4a-0069-4af3-b76b-6fd0be528cda}'
HV_FCOPY_GUID = '{34d14be3-dee4-41c8-9ae7-6b174977c192}'
HV_KBD_GUID = '{f912ad6d-2b17-48ea-bd65-f927a61c7684}'
REMOTE_DESKTOP_GUID = '{276aacf4-ac15-426c-98dd-7521ad3f01fe}'
VOLUME_SHADOW_COPY_GUID = '{35fa2e29-ea23-4236-96ae-3a6ebacba440}'
AVMA_GUID = '{3375baf4-9e15-4b30-b765-67acb10d607b}'
DESKTOP_CONTROL_GUID = '{f8e65716-3cb3-4a06-9a60-1889c5cccab5}'

hv_guid_desc = {
    HV_NIC_GUID: 'Microsoft Hyper-V Network',
    HV_IDE_GUID: 'Microsoft Hyper-V IDE',
    HV_SCSI_GUID: 'Microsoft Hyper-V SCSI Controller',
    HV_SHUTDOWN_GUID: 'Microsoft Hyper-V Shutdown',
    HV_TS_GUID: 'Microsoft Hyper-V Time Synch',
    HV_HEART_BEAT_GUID: 'Microsoft Hyper-V Heartbeat',
    HV_KVP_GUID: 'Microsoft Hyper-V KVP',
    HV_DM_GUID: 'Microsoft Hyper-V Dynamic memory',
    HV_MOUSE_GUID: 'Microsoft Hyper-V Mouse',
    HV_VSS_GUID: 'Microsoft Hyper-V VSS Backup/Restore',
    HV_SYNTHVID_GUID: 'Microsoft Hyper-V Synthetic Video',
    HV_SYNTHFC_GUID: 'Microsoft Hyper-V Synthetic FC',
    HV_FCOPY_GUID: 'Microsoft Hyper-V Guest File Copy Service',
    HV_KBD_GUID: 'Microsoft Hyper-V Virtual Keyboard',
    REMOTE_DESKTOP_GUID: 'Microsoft Hyper-V Remote Desktop Virtualization',
    VOLUME_SHADOW_COPY_GUID: 'Microsoft Hyper-V Volume Shadow Copy',
    AVMA_GUID: 'Microsoft Hyper-V Automatic Virtual Machine Activation (AVMA)',
    DESKTOP_CONTROL_GUID: 'Microsoft Hyper-V Remote Desktop Control Channel'
}

### VMBUS PACKET DEFINES #######################################################

vm_pkt = {
    0x0: 'VM_PKT_INVALID',
    0x1: 'VM_PKT_SYNCH',
    0x2: 'VM_PKT_ADD_XFER_PAGESET',
    0x3: 'VM_PKT_RM_XFER_PAGESET',
    0x4: 'VM_PKT_ESTABLISH_GPADL',
    0x5: 'VM_PKT_TEARDOWN_GPADL',
    0x6: 'VM_PKT_DATA_INBAND',
    0x7: 'VM_PKT_DATA_USING_XFER_PAGES',
    0x8: 'VM_PKT_DATA_USING_GPADL',
    0x9: 'VM_PKT_DATA_USING_GPA_DIRECT',
    0xa: 'VM_PKT_CANCEL_REQUEST',
    0xb: 'VM_PKT_COMP',
    0xc: 'VM_PKT_DATA_USING_ADDITIONAL_PKT',
    0xd: 'VM_PKT_ADDITIONAL_DATA'
}

VMBUS_DATA_PACKET_FLAG_COMPLETION_REQUESTED = 1

################################################################################


def set_variables(varlist):
    for i in varlist:
        var = re.sub(r"([a-z])([A-Z]+)", r"\1_\2", varlist[i])
        var = var.upper()
        exec(f'global {var}; {var}={i:d}')


set_variables(msrs)
set_variables(hypercall_status_codes)
set_variables(hypercall_names)
set_variables(hv_porttype)
set_variables(hv_message_type)
set_variables(channel_flags)
set_variables(vm_pkt)



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


"""
Compatible SMM memory (SMRAM) Protection check module
This CHIPSEC module simply reads SMRAMC and checks that D_LCK is set.

Reference:
In 2006, `Security Issues Related to Pentium System Management Mode <http://www.ssi.gouv.fr/archive/fr/sciences/fichiers/lti/cansecwest2006-duflot.pdf>`_ outlined a configuration issue where compatibility SMRAM was not locked on some platforms. This means that ring 0 software was able to modify System Management Mode (SMM) code and data that should have been protected.

In Compatability SMRAM (CSEG), access to memory is defined by the SMRAMC register. When SMRAMC[D_LCK] is not set by the BIOS, SMRAM can be accessed even when the CPU is not in SMM. Such attacks were also described in `Using CPU SMM to Circumvent OS Security Functions <http://fawlty.cs.usfca.edu/~cruse/cs630f06/duflot.pdf>`_ and `Using SMM for Other Purposes <http://phrack.org/issues/65/7.html>`_.

usage:
    ``chipsec_main -m common.smm``

Examples:
    >>> chipsec_main.py -m common.smm

This module will only run on client (core) platforms that have PCI0.0.0_SMRAMC defined.
"""

from chipsec.module_common import BaseModule, BIOS, SMM
from chipsec.library.returncode import ModuleResult
from typing import List

TAGS = [BIOS, SMM]
METADATA_TAGS = ['OPENSOURCE', 'IA', 'COMMON', 'SMM']


class smm(BaseModule):

    def __init__(self):
        BaseModule.__init__(self)

    def is_supported(self) -> bool:
        if self.cs.is_core() and self.cs.register.is_defined('PCI0.0.0_SMRAMC'):
            return True
        self.logger.log("Either not a Core (client) platform or 'PCI0.0.0_SMRAMC' not defined for platform. Skipping module.")
        return False

    def check_SMRAMC(self) -> int:

        regval = self.cs.register.read('PCI0.0.0_SMRAMC')
        g_smrame = self.cs.register.get_field('PCI0.0.0_SMRAMC', regval, 'G_SMRAME')
        d_open = self.cs.register.get_field('PCI0.0.0_SMRAMC', regval, 'D_OPEN')
        d_lock = self.cs.register.get_field('PCI0.0.0_SMRAMC', regval, 'D_LCK')

        self.cs.register.print('PCI0.0.0_SMRAMC', regval)

        if 1 == g_smrame:
            self.logger.log('[*] Compatible SMRAM is enabled')
            if (1 == d_lock) and (0 == d_open):
                res = ModuleResult.PASSED
                self.logger.log_passed('Compatible SMRAM is locked down')
            else:
                res = ModuleResult.FAILED
                self.logger.log_failed('Compatible SMRAM is not properly locked. Expected ( D_LCK = 1, D_OPEN = 0 )')
                self.result.setStatusBit(self.result.status.LOCKS)
        else:
            res = ModuleResult.NOTAPPLICABLE
            self.result.setStatusBit(self.result.status.FEATURE_DISABLED)
            self.logger.log('[*] Compatible SMRAM is not enabled. Skipping..')

        return self.result.getReturnCode(res)

    def run(self, module_argv: List[str]) -> int:
        self.logger.start_test('Compatible SMM memory (SMRAM) Protection')
        self.res = self.check_SMRAMC()
        return self.res



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


from typing import Dict
from chipsec.library.tpm import tpm2_commands
from chipsec.library.tpm import tpm1_commands


# Common defines
TPM1 = '1'
TPM2 = '2'
TPM_CRB = 'crb'
TPM_FIFO = 'fifo'
TPM_FIFO_LEGACY = 'fifo_legacy'

TPM2_INTERFACE_ADDR = 0xFED40030

COMMAND_FORMAT = "=HIIIII"

COMMANDREADY = 0x40
TPMGO = 0x20
HEADERSIZE = 0x0A
HEADERFORMAT = '>HII'
BEENSEIZED  = 0x10
REQUESTUSE  = 0x2
ACTIVELOCALITY = 0x20
DATAAVAIL = 0x10

TPM_DATAFIFO = 0x0024
TPM_STS = 0x0018
TPM_DIDVID = 0x0F00
TPM_ACCESS = 0x0000
TPM_RID = 0x0F04
TPM_INTCAP = 0x0014
TPM_INTENABLE = 0x0008

list_of_registers = ['TPM_ACCESS',
                    'TPM_STS',
                    'TPM_DID_VID',
                    'TPM_RID',
                    'TPM_INTF_CAPABILITY',
                    'TPM_INT_ENABLE']

LOCALITY: Dict[str, int] = {
    '0': 0x0000,
    '1': 0x1000,
    '2': 0x2000,
    '3': 0x3000,
    '4': 0x4000
}

PCR: Dict[int, int] = {
    0: 0x00000000,
    1: 0x01000000,
    2: 0x02000000,
    3: 0x03000000,
    4: 0x04000000,
    5: 0x05000000,
    6: 0x06000000,
    7: 0x07000000,
    8: 0x08000000,
    9: 0x09000000,
    10: 0x0a000000,
    11: 0x0b000000,
    12: 0x0c000000,
    13: 0x0d000000,
    14: 0x0e000000,
    15: 0x0f000000,
    16: 0x10000000,
    17: 0x11000000,
    18: 0x12000000,
    19: 0x13000000,
    20: 0x14000000,
    21: 0x15000000,
    22: 0x16000000,
    23: 0x17000000,
    24: 0x18000000,
    25: 0x19000000,
    26: 0x1a000000,
    27: 0x1b000000,
    28: 0x1c000000,
    29: 0x1d000000,
    30: 0x1e000000
}

vendor_ids = {
    0x1022: 'AMD',
    0x6688: 'Ant Group',
    0x1114: 'Atmel',
    0x14E4: 'Broadcom',
    0xC5C0: 'Cisco',
    0x232B: 'FlySlice Technologies',
    0x232A: 'Fuzhou Rockchip',
    0x6666: 'Google',
    0x103C: 'HPI',
    0x1590: 'HPE',
    0x8888: 'Huawei',
    0x1014: 'IBM',
    0x15D1: 'Infineon',
    0x8086: 'Intel',
    0x17AA: 'Lenovo',
    0x1414: 'Microsoft',
    0x100B: 'National Semi',
    0x1B4E: 'Nationz',
    0x9999: 'NSING',
    0x1050: 'Nuvoton Technology1',
    0x1011: 'Qualcomm',
    0x144D: 'Samsung',
    0x5ECE: 'SecEdge',
    0x19FA: 'Sinosun',
    0x1055: 'SMSC',
    0x025E: 'Solidigm',
    0x104A: 'STMicroelectronics',
    0x104C: 'Texas Instruments',
    0x2406: 'Wisekey'
}


# TPM 1.2 defines
TPM_TAG_RQU_COMMAND = 0xc100
TPM_TAG_RQU_AUTH1_COMMAND = 0xc200
TPM_TAG_RQU_AUTH2_COMMAND = 0xC300
TPM_TAG_RSP_COMMAND = 0xC400
TPM_TAG_RSP_AUTH1_COMMAND = 0xC500
TPM_TAG_RSP_AUTH2_COMMAND = 0xC600

TPM_ORD_CONTINUESELFTEST = 0x53000000
TPM_ORD_FORCECLEAR = 0x5D000000
TPM_ORD_GETCAPABILITY = 0x65000000
TPM_ORD_NV_DEFINESPACE = 0xCC000000
TPM_ORD_NV_READVALUE = 0xCF000000
TPM_ORD_NV_WRITEVALUE = 0xCD000000
TPM_ORD_PCRREAD = 0x15000000
TPM_ORD_PHYSICALDISABLE = 0x70000000
TPM_ORD_PHYSICALENABLE = 0x6F000000
TPM_ORD_PHYSICALSETDEACTIVATED = 0x72000000
TPM_ORD_STARTUP = 0x99000000
TPM_ORD_SAVESTATE = 0x98000000
TSC_ORD_PHYSICALPRESENCE = 0x0A000040
TSC_ORD_RESETESTABLISHMENTBIT = 0x0B000040

COMMANDS_12: Dict[str, callable] = {
        'pcrread': tpm1_commands.pcrread,
        'nvread': tpm1_commands.nvread,
        'startup': tpm1_commands.startup,
        'continueselftest': tpm1_commands.continueselftest,
        'forceclear': tpm1_commands.forceclear,
        'header': tpm1_commands.TPM_RESPONSE_HEADER
    }

STARTUP: Dict[int, int] = {
    1: 0x0100,
    2: 0x0200,
    3: 0x0300
}

STATUS: Dict[int, str] = {
    0x00: 'Success',
    0x01: 'ERROR: Authentication Failed',
    0x02: 'ERROR: The index to a PCR, DIR or other register is incorrect',
    0x03: 'ERROR: One or more parameter is bad',
    0x04: 'ERROR: An operation completed successfully but the auditing of that operation failed',
    0x05: 'ERROR: The clear disable flag is set and all clear operations now require physical access',
    0x06: 'ERROR: The TPM is deactivated',
    0x07: 'ERROR: The TPM is disabled',
    0x08: 'ERROR: The target command has been disabled',
    0x09: 'ERROR: The operation failed',
    0x0A: 'ERROR: The ordinal was unknown or inconsistent',
    0x0B: 'ERROR: The ability to install an owner is disabled',
    0x0C: 'ERROR: The key handle can not be interpreted',
    0x0D: 'ERROR: The key handle points to an invalid key',
    0x0E: 'ERROR: Unacceptable encryption scheme',
    0x0F: 'ERROR: Migration authorization failed',
    0x10: 'ERROR: PCR information could not be interpreted',
    0x11: 'ERROR: No room to load key',
    0x12: 'ERROR: There is no SRK set',
    0x13: 'ERROR: An encrypted blob is invalid or was not created by this TPM',
    0x14: 'ERROR: There is already an Owner',
    0x15: 'ERROR: The TPM has insufficient internal resources to perform the requested action',
    0x16: 'ERROR: A random string was too short',
    0x17: 'ERROR: The TPM does not have the space to perform the operation',
    0x18: 'ERROR: The named PCR value does not match the current PCR value.',
    0x19: 'ERROR: The paramSize argument to the command has the incorrect value',
    0x1A: 'ERROR: There is no existing SHA-1 thread.',
    0x1B: 'ERROR: The calculation is unable to proceed because the existing SHA-1 thread has already encountered an error',
    0x1C: 'ERROR: Self-test has failed and the TPM has shut-down',
    0x1D: 'ERROR: The authorization for the second key in a 2 key function failed authorization',
    0x1E: 'ERROR: The tag value sent to for a command is invalid',
    0x1F: 'ERROR: An IO error occurred transmitting information to the TPM',
    0x20: 'ERROR: The encryption process had a problem',
    0x21: 'ERROR: The decryption process did not complete',
    0x22: 'ERROR: An invalid handle was used',
    0x23: 'ERROR: The TPM does not a EK installed',
    0x24: 'ERROR: The usage of a key is not allowed',
    0x25: 'ERROR: The submitted entity type is not allowed',
    0x26: 'ERROR: The command was received in the wrong sequence relative to TPM_Init and a subsequent TPM_Startup',
    0x27: 'ERROR: Signed data cannot include additional DER information',
    0x28: 'ERROR: The key properties in TPM_KEY_PARMs are not supported by this TPM',
    0x29: 'ERROR: The migration properties of this key are incorrect',
    0x2A: 'ERROR: The signature or encryption scheme for this key is incorrect or not permitted in this situation',
    0x2B: 'ERROR: The size of the data (or blob) parameter is bad or inconsistent with the referenced key',
    0x2C: 'ERROR: A parameter is bad',
    0x2D: 'ERROR: Either the physicalPresence or physicalPresenceLock bits have the wrong value',
    0x2E: 'ERROR: The TPM cannot perform this version of the capability',
    0x2F: 'ERROR: The TPM does not allow for wrapped transport sessions',
    0x30: 'ERROR: TPM audit construction failed and the underlying command was returning a failure code also',
    0x31: 'ERROR: TPM audit construction failed and the underlying command was returning success',
    0x32: 'ERROR: Attempt to reset a PCR register that does not have the resettable attribute',
    0x33: 'ERROR: Attempt to reset a PCR register that requires locality and locality modifier not part of command transport',
    0x34: 'ERROR: Make identity blob not properly typed',
    0x35: 'ERROR: When saving context identified resource type does not match actual resource',
    0x36: 'ERROR: The TPM is attempting to execute a command only available when in FIPS mode',
    0x37: 'ERROR: The command is attempting to use an invalid family ID',
    0x38: 'ERROR: The permission to manipulate the NV storage is not available',
    0x39: 'ERROR: The operation requires a signed command',
    0x3A: 'ERROR: Wrong operation to load an NV key',
    0x3B: 'ERROR: NV_LoadKey blob requires both owner and blob authorization',
    0x3C: 'ERROR: The NV area is locked and not writeable',
    0x3D: 'ERROR: The locality is incorrect for the attempted operation',
    0x3E: 'ERROR: The NV area is read only and can?t be written to',
    0x3F: 'ERROR: There is no protection on the write to the NV area',
    0x40: 'ERROR: The family count value does not match',
    0x41: 'ERROR: The NV area has already been written to',
    0x42: 'ERROR: The NV area attributes conflict',
    0x43: 'ERROR: The structure tag and version are invalid or inconsistent',
    0x44: 'ERROR: The key is under control of the TPM Owner and can only be evicted by the TPM Owner',
    0x45: 'ERROR: The counter handle is incorrect',
    0x46: 'ERROR: The write is not a complete write of the area',
    0x47: 'ERROR: The gap between saved context counts is too large',
    0x48: 'ERROR: The maximum number of NV writes without an owner has been exceeded',
    0x49: 'ERROR: No operator AuthData value is set',
    0x4A: 'ERROR: The resource pointed to by context is not loaded',
    0x4B: 'ERROR: The delegate administration is locked',
    0x4C: 'ERROR: Attempt to manage a family other then the delegated family',
    0x4D: 'ERROR: Delegation table management not enabled',
    0x4E: 'ERROR: There was a command executed outside of an exclusive transport session',
    0x4F: 'ERROR: Attempt to context save a owner evict controlled key',
    0x50: 'ERROR: The DAA command has no resources available to execute the command',
    0x51: 'ERROR: The consistency check on DAA parameter inputData0 has failed',
    0x52: 'ERROR: The consistency check on DAA parameter inputData1 has failed',
    0x53: 'ERROR: The consistency check on DAA_issuerSettings has failed',
    0x54: 'ERROR: The consistency check on DAA_tpmSpecific has failed',
    0x55: 'ERROR: The atomic process indicated by the submitted DAA command is not the expected process',
    0x56: 'ERROR: The issuers validity check has detected an inconsistency',
    0x57: 'ERROR: The consistency check on w has failed',
    0x58: 'ERROR: The handle is incorrect',
    0x59: 'ERROR: Delegation is not correct',
    0x5A: 'ERROR: The context blob is invalid',
    0x5B: 'ERROR: Too many contexts held by the TPM',
    0x5C: 'ERROR: Migration authority signature validation failure',
    0x5D: 'ERROR: Migration destination not authenticated',
    0x5E: 'ERROR: Migration source incorrect',
    0x5F: 'ERROR: Incorrect migration authority',
    0x60: 'ERROR: TBD',
    0x61: 'ERROR: Attempt to revoke the EK and the EK is not revocable',
    0x62: 'ERROR: Bad signature of CMK ticket',
    0x63: 'ERROR: There is no room in the context list for additional contexts',
    0x800: 'NON-FATAL ERROR: The TPM is too busy to respond to the command immediately, but the command could be resubmitted at a later time',
    0x801: 'NON-FATAL ERROR: TPM_ContinueSelfTest has not been run.',
    0x802: 'NON-FATAL ERROR: The TPM is currently executing the actions of TPM_ContinueSelfTest because the ordinal required resources that have not been tested',
    0x803: 'NON-FATAL ERROR: The TPM is defending against dictionary attacks and is in some time-out period.'
}

STARTUP: Dict[int, int] = {
    1: 0x0100,
    2: 0x0200,
    3: 0x0300
}


# TPM 2.0 defines
TPM2_NVUNDEFINESPACE_COMMAND = 0x22010000
TPM2_NVDEFINESPACE_COMMAND = 0x2A010000
TPM2_NVWRITE_COMMAND = 0x37010000
TPM2_SELFTEST_COMMAND = 0x43010000
TPM2_STARTUP_COMMAND = 0x44010000
TPM2_SHUTDOWN_COMMAND = 0x45010000
TPM2_NVREAD_COMMAND = 0x4E010000
TPM2_PCRREAD_COMMAND = 0x7E010000

COMMANDS_20 = {
        'startup': tpm2_commands.startup,
        'shutdown': tpm2_commands.shutdown,
        'selftest': tpm2_commands.selftest,
        'nvread': tpm2_commands.nvread,
        'nvwrite': tpm2_commands.nvwrite,
        'pcrread': tpm2_commands.pcrread,
        'nvdefinespace': tpm2_commands.nvdefinespace,
        'nvundefinespace': tpm2_commands.nvundefinespace,
        'header': tpm2_commands.TPM_RESPONSE_HEADER
    }

TPM_ST_NO_SESSIONS = 0x0180
TPM_ST_SESSIONS = 0x0280

SESSIONS: Dict[int, int] = {
    0: TPM_ST_NO_SESSIONS,
    1: TPM_ST_SESSIONS
}

TPM_SU: Dict[int, int] = {
    0: 0x0000,  # TPM_SU_CLEAR
    1: 0x0001   # TPM_SU_STATE
}

TPMI_YES_NO: Dict[int, int] = {
    0: 0x00,
    1: 0x01
}

TPMI_RH_NV_AUTH: Dict[int, int] = {
    1: 0x4000000C,  # TPM_RH_PLATFORM
    2: 0x40000001   # TPM_RH_OWNER
}

NV_INDEX_FIRST = 0x01 << 24
NV_INDEX_LAST = NV_INDEX_FIRST + 0x00FFFFFF

RESPONSE_TAG = {}

RESPONSE_CODE = {
    # https://trustedcomputinggroup.org/wp-content/uploads/TPM-Rev-2.0-Part-2-Structures-01.38.pdf
    # Section 6.6.3
    # VER1
    0x000: ['TPM_RC_SUCCESS', 'No error.'],
    0x01E: ['TPM_RC_BAD_TAG', 'Bad tag in the message.'],
    0x100: ['TPM_RC_INITIALIZE', 'TPM not initialized by TPM2_Startup() or already initialized.'],
    0x101: ['TPM_RC_FAILURE', 'Commands not being accepted because of a TPM failure (can be returned by TPM2_GetTestResult() as testResult).'],
    0x103: ['TPM_RC_SEQUENCE', 'Improper use of a sequence handle.'],
    0x10B: ['TPM_RC_PRIVATE', 'Not currently used.'],
    0x119: ['TPM_RC_HMAC', 'Not currently used.'],
    0x120: ['TPM_RC_DISABLED', 'The command is disabled.'],
    0x121: ['TPM_RC_EXCLUSIVE', 'Command failed because audit sequence required exclusivity.'],
    0x124: ['TPM_RC_AUTH_TYPE', 'Authorization handle is not correct for command.'],
    0x125: ['TPM_RC_AUTH_MISSING', 'Command requires an authorization session for handle and it is not present.'],
    0x126: ['TPM_RC_POLICY', 'Policy failure in math operation or invalid authPolicy value.'],
    0x127: ['TPM_RC_PCR', 'PCR check failed.'],
    0x128: ['TPM_RC_PCR_CHANGED', 'PCR has changed since last checked.'],
    0x12D: ['TPM_RC_UPGRADE', 'Indicates that the TPM is in field upgrade mode (except for TPM2_FieldUpgradeData() command).'],
    0x12E: ['TPM_RC_TOO_MANY_CONTEXTS', 'Context ID counter is at maximum.'],
    0x12F: ['TPM_RC_AUTH_UNAVAILABLE', 'authValue or authPolicy is not available for selected entity.'],
    0x130: ['TPM_RC_REBOOT', 'TPM_Init and Startup(CLEAR) required before TPM can resume operation.'],
    0x131: ['TPM_RC_UNBALANCED', 'Protection algorithms (hash and symmetric) are not reasonably balanced (hash digest size must be larger than key size of symmetric algorithm).'],
    0x142: ['TPM_RC_COMMAND_SIZE', 'Command commandSize value inconsistent with contents of command buffer or insufficient size for header.'],
    0x143: ['TPM_RC_COMMAND_CODE', 'Command code not supported.'],
    0x144: ['TPM_RC_AUTHSIZE', 'AuthorizationSize value is out of range or too many octets in Authorization Area.'],
    0x145: ['TPM_RC_AUTH_CONTEXT', 'Use of authorization session with context or another incompatible command.'],
    0x146: ['TPM_RC_NV_RANGE', 'NV offset + size is out of range.'],
    0x147: ['TPM_RC_NV_SIZE', 'Requested allocation size is too large.'],
    0x148: ['TPM_RC_NV_LOCKED', 'NV access locked.'],
    0x149: ['TPM_RC_NV_AUTHORIZATION', 'NV access authorization fails.'],
    0x14A: ['TPM_RC_NV_UNINITIALIZED', 'NV Index used before initialization or state from TPM2_Shutdown(STATE) cannot be restored.'],
    0x14B: ['TPM_RC_NV_SPACE', 'Insufficient space for NV allocation.'],
    0x14C: ['TPM_RC_NV_DEFINED', 'NV Index or persistent object already defined.'],
    0x150: ['TPM_RC_BAD_CONTEXT', 'Invalid context in TPM2_ContextLoad().'],
    0x151: ['TPM_RC_CPHASH', 'cpHash value already set or incorrect for use.'],
    0x152: ['TPM_RC_PARENT', 'Invalid parent handle.'],
    0x153: ['TPM_RC_NEEDS_TEST', 'Function needs testing.'],
    0x154: ['TPM_RC_NO_RESULT', 'Returned when an internal function cannot process a request due to unspecified problem. Usually due to invalid parameters.'],
    0x155: ['TPM_RC_SENSITIVE', 'Sensitive area did not unmarshal correctly after decryption.'],
    0x156: ['TPM_RC_READ_ONLY', 'Command failed because TPM is in Read-Only mode.'],
    0x17F: ['TPM_RC_MAX_FM0', 'Largest version 1 code that is not a warning.'],
    # FMT1
    0x081: ['TPM_RC_ASYMMETRIC', 'Asymmetric algorithm not supported or incorrect.'],
    0x082: ['TPM_RC_ATTRIBUTES', 'Inconsistent attributes.'],
    0x083: ['TPM_RC_HASH', 'Unsupported or inappropriate hash algorithm.'],
    0x084: ['TPM_RC_VALUE', 'Value out of range or not correct for context.'],
    0x085: ['TPM_RC_HIERARCHY', 'Hierarchy is not enabled or incorrect for use.'],
    0x087: ['TPM_RC_KEY_SIZE', 'Key size not supported.'],
    0x088: ['TPM_RC_MGF', 'Mask generation function not supported.'],
    0x089: ['TPM_RC_MODE', 'Mode of operation not supported.'],
    0x08A: ['TPM_RC_TYPE', 'Value type is not appropriate.'],
    0x08B: ['TPM_RC_HANDLE', 'Handle not correct for use.'],
    0x08C: ['TPM_RC_KDF', 'Unsupported or inappropriate key derivation function.'],
    0x08D: ['TPM_RC_RANGE', 'Value out of allowed range.'],
    0x08E: ['TPM_RC_AUTH_FAIL', 'Authorization HMAC check failed or DA counter incremented.'],
    0x08F: ['TPM_RC_NONCE', 'Invalid nonce size or nonce value mismatch.'],
    0x090: ['TPM_RC_PP', 'Authorization requires assertion of PP.'],
    0x092: ['TPM_RC_SCHEME', 'Unsupported or incompatible scheme.'],
    0x095: ['TPM_RC_SIZE', 'Structure is the wrong size.'],
    0x096: ['TPM_RC_SYMMETRIC', 'Unsupported symmetric algorithm or key size.'],
    0x097: ['TPM_RC_TAG', 'Incorrect structure tag.'],
    0x098: ['TPM_RC_SELECTOR', 'Union selector is incorrect.'],
    0x09A: ['TPM_RC_INSUFFICIENT', 'TPM unable to unmarshal due to insufficient octets in buffer.'],
    0x09B: ['TPM_RC_SIGNATURE', 'Signature is invalid.'],
    0x09C: ['TPM_RC_KEY', 'Key fields not compatible with selected use.'],
    0x09D: ['TPM_RC_POLICY_FAIL', 'Policy check failed.'],
    0x09F: ['TPM_RC_INTEGRITY', 'Integrity check failed.'],
    0x0A0: ['TPM_RC_TICKET', 'Invalid ticket.'],
    0x0A1: ['TPM_RC_RESERVED_BITS', 'Reserved bits not set to zero as required.'],
    0x0A2: ['TPM_RC_BAD_AUTH', 'Authorization failure without DA implications.'],
    0x0A3: ['TPM_RC_EXPIRED', 'The policy has expired.'],
    0x0A4: ['TPM_RC_POLICY_CC', 'The commandCode in the policy is not the commandCode of the command.'],
    0x0A5: ['TPM_RC_BINDING', 'Public and sensitive portions of an object are not cryptographically bound.'],
    0x0A6: ['TPM_RC_CURVE', 'Curve not supported.'],
    0x0A7: ['TPM_RC_ECC_POINT', 'Point is not on the required curve.'],
    0x0A8: ['TPM_RC_FW_LIMITED', 'The hierarchy is firmware-limited but the Firmware Secret is unavailable.'],
    0x0A9: ['TPM_RC_SVN_LIMITED', 'The hierarchy is SVN-limited but the Firmware SVN Secret associated with the given SVN is unavailable.'],
    0x0AA: ['TPM_RC_CHANNEL', 'Command requires secure channel protection.'],
    0x0AB: ['TPM_RC_CHANNEL_KEY', 'Secure channel was not established with required requester or TPM key.'],
    # Warnings
    0x901: ['TPM_RC_CONTEXT_GAP', 'Gap for context ID is too large.'],
    0x902: ['TPM_RC_OBJECT_MEMORY', 'Out of memory for object contexts'],
    0x903: ['TPM_RC_SESSION_MEMORY', 'Out of memory for session contexts'],
    0x904: ['TPM_RC_MEMORY', 'Out of shared object/session memory or need space for internal operations'],
    0x905: ['TPM_RC_SESSION_HANDLES', 'Out of session handles - a session must be flushed before a new session may be created'],
    0x906: ['TPM_RC_OBJECT_HANDLES', 'Out of object handles - the handle space for objects is depleted and a reboot is required'],
    0x907: ['TPM_RC_LOCALITY', 'Bad locality'],
    0x908: ['TPM_RC_YIELDED', 'The TPM has suspended operation on the command; forward progress was made and the command may be retried'],
    0x909: ['TPM_RC_CANCELED', 'The command was canceled'],
    0x90A: ['TPM_RC_TESTING', 'TPM is performing self-tests'],
    0x910: ['TPM_RC_REFERENCE_H0', 'The 1st handle in the handle area references a transient object or session that is not loaded'],
    0x911: ['TPM_RC_REFERENCE_H1', 'The 2nd handle in the handle area references a transient object or session that is not loaded'],
    0x912: ['TPM_RC_REFERENCE_H2', 'The 3rd handle in the handle area references a transient object or session that is not loaded'],
    0x913: ['TPM_RC_REFERENCE_H3', 'The 4th handle in the handle area references a transient object or session that is not loaded'],
    0x914: ['TPM_RC_REFERENCE_H4', 'The 5th handle in the handle area references a transient object or session that is not loaded'],
    0x915: ['TPM_RC_REFERENCE_H5', 'The 6th handle in the handle area references a transient object or session that is not loaded'],
    0x916: ['TPM_RC_REFERENCE_H6', 'The 7th handle in the handle area references a transient object or session that is not loaded'],
    0x918: ['TPM_RC_REFERENCE_S0', 'The 1st authorization session handle references a session that is not loaded'],
    0x919: ['TPM_RC_REFERENCE_S1', 'The 2nd authorization session handle references a session that is not loaded'],
    0x91A: ['TPM_RC_REFERENCE_S2', 'The 3rd authorization session handle references a session that is not loaded'],
    0x91B: ['TPM_RC_REFERENCE_S3', 'The 4th authorization session handle references a session that is not loaded'],
    0x91C: ['TPM_RC_REFERENCE_S4', 'The 5th authorization session handle references a session that is not loaded'],
    0x91D: ['TPM_RC_REFERENCE_S5', 'The 6th authorization session handle references a session that is not loaded'],
    0x91E: ['TPM_RC_REFERENCE_S6', 'The 7th authorization session handle references a session that is not loaded'],
    0x920: ['TPM_RC_NV_RATE', 'The TPM is rate-limiting accesses to prevent wearout of NV'],
    0x921: ['TPM_RC_LOCKOUT', 'Authorizations for objects subject to DA protection are not allowed at this time because the TPM is in DA lockout mode'],
    0x922: ['TPM_RC_RETRY', 'The TPM was not able to start the command'],
    0x923: ['TPM_RC_NV_UNAVAILABLE', 'The command may require writing of NV and NV is not currently accessible'],
    0x97F: ['TPM_RC_NOT_USED', 'This value is reserved and shall not be returned by the TPM'],
    # Additional defines - Offsets, not actual return codes.
    # 0x000: ['TPM_RC_H', 'Add to a handle-related error.'],
    # 0x040: ['TPM_RC_P', 'Add to a parameter-related error.'],
    # 0x800: ['TPM_RC_S', 'Add to a session-related error.'],
    # 0x100: ['TPM_RC_1', 'Add to a parameter-, handle-, or session-related error.'],
    # 0x200: ['TPM_RC_2', 'Add to a parameter-, handle-, or session-related error.'],
    # 0x300: ['TPM_RC_3', 'Add to a parameter-, handle-, or session-related error.'],
    # 0x400: ['TPM_RC_4', 'Add to a parameter-, handle-, or session-related error.'],
    # 0x500: ['TPM_RC_5', 'Add to a parameter-, handle-, or session-related error.'],
    # 0x600: ['TPM_RC_6', 'Add to a parameter-, handle-, or session-related error.'],
    # 0x700: ['TPM_RC_7', 'Add to a parameter-, handle-, or session-related error.'],
    # 0x800: ['TPM_RC_8', 'Add to a parameter-related error.'],
    # 0x900: ['TPM_RC_9', 'Add to a parameter-related error.'],
    # 0xA00: ['TPM_RC_A', 'Add to a parameter-related error.'],
    # 0xB00: ['TPM_RC_B', 'Add to a parameter-related error.'],
    # 0xC00: ['TPM_RC_C', 'Add to a parameter-related error.'],
    # 0xD00: ['TPM_RC_D', 'Add to a parameter-related error.'],
    # 0xE00: ['TPM_RC_E', 'Add to a parameter-related error.'],
    # 0xF00: ['TPM_RC_F', 'Add to a parameter-related error.'],
    # 0xF00: ['TPM_RC_N_MASK', 'Number mask.'],
}


#!/usr/bin/env python3
# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2023, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

"""
Script to automatically generate CHIPSEC's documentation using Sphinx.
It generates PDF plus either HTML or JSON formats.

Usage:
    ``python3 create_manual.py [format]``

    ``format`` - html or json

Examples:
    >>> python3 create_manual.py
    >>> python3 create_manual.py html
    >>> python3 create_manual.py json

References:
https://www.sphinx-doc.org/en/master/man/sphinx-apidoc.html
https://www.sphinx-doc.org/en/master/man/sphinx-build.html
https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html
"""

import os
try:
    from collections.abc import Sequence
except ImportError:
    from typing import Sequence
    print('!! Unable to import collections.abc !!')
import shutil
import sys

DOCS_DIR = os.getcwd()
RM_DIR = os.path.join(DOCS_DIR, '_remove')
SPHINX_DIR = os.path.join(DOCS_DIR, 'sphinx')
SPHINX_MOD_DIR = os.path.join(SPHINX_DIR, 'modules')
SPHINX_SCRIPTS_DIR = os.path.join(SPHINX_DIR, '_scripts')
CHIPSEC_DIR = os.path.normpath(DOCS_DIR + os.sep + os.pardir)


def RunAutoDoc() -> None:
    try:
        os.system(f'sphinx-apidoc -e -f -T -d 10 -o modules {CHIPSEC_DIR} {os.path.join(CHIPSEC_DIR, "*test*")} {os.path.join(CHIPSEC_DIR, "*exceptions*")} {os.path.join(CHIPSEC_DIR, "*tool*")}')
    except Exception:
        print('Unable to run sphinx-apidoc')
        raise


def CleanupFilesNotWantedInDoc() -> None:
    NotWantedFilesList = []
    for file in os.listdir(RM_DIR):
        with open(os.path.join(RM_DIR, file), 'r') as f:
            NotWantedFilesList = f.read()
        for not_needed_file in NotWantedFilesList.split(','):
            RemoveFile(not_needed_file)


def RemoveFile(file):
    try:
        os.remove(os.path.join(SPHINX_MOD_DIR, file))
    except Exception:
        print(f'\t\tUnable to remove {file}!!!')


def RunScripts() -> None:
    for script in os.listdir(os.path.join(SPHINX_SCRIPTS_DIR)):
        try:
            os.system(f'python3 {os.path.join(SPHINX_SCRIPTS_DIR, script)}')
        except Exception:
            print(f'Unable to run script: {script}')
            raise


def GeneratePDF() -> None:
    try:
        os.system(f'sphinx-build -b pdf -T {SPHINX_DIR} {CHIPSEC_DIR}')
    except Exception:
        print('Unable to generate PDF')
        raise


def GenerateHTML() -> None:
    try:
        os.system(f'sphinx-build -b html -T {SPHINX_DIR} {os.path.join(CHIPSEC_DIR, "manual")}')
    except Exception:
        print('Unable to generate HTML')
        raise


def GenerateJSON() -> None:
    try:
        os.system(f'sphinx-build -b json -T {SPHINX_DIR} {os.path.join(CHIPSEC_DIR, "manualJson")}')
    except Exception:
        print('Unable to generate JSON')
        raise


format_options_functions = {
    'html': GenerateHTML,
    'json': GenerateJSON
}


def GenerateHTMLorJSON(option: str) -> None:
    try:
        format_options_functions[option]()
    except Exception:
        print('Invalid format option')
        raise


def DeleteSphinxCollateral() -> None:
    shutil.rmtree(os.path.join(CHIPSEC_DIR, '.doctrees'))
    shutil.rmtree(os.path.join(SPHINX_DIR, 'logs'))
    shutil.rmtree(SPHINX_MOD_DIR)


def main(argv: Sequence[str] = sys.argv[1:]):
    print('******************** BUILDING DOCUMENTATION **************************')
    os.chdir(SPHINX_DIR)
    try:
        RunAutoDoc()
        CleanupFilesNotWantedInDoc()
        RunScripts()
        if (len(argv)==1):
            GenerateHTMLorJSON(argv[0])
        GeneratePDF()
        DeleteSphinxCollateral()
    except Exception:
        return 1
    return 0


if __name__ == "__main__":
    sys.exit(main())
